#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import getopt
import io
import json
import logging
import os
import sys
import time
import zipfile

from ai_flow.ai_graph.ai_graph import AIGraph
from ai_flow.ai_graph.ai_node import AINode, ReadDatasetNode, WriteDatasetNode
from ai_flow.ai_graph.data_edge import DataEdge
from ai_flow.endpoint.server.server_config import AIFlowServerConfig
from ai_flow.meta.workflow_meta import WorkflowMeta
from ai_flow.plugin_interface.scheduler_interface import Scheduler, SchedulerFactory
from ai_flow.scheduler_service.service.config import SchedulerServiceConfig
from ai_flow.settings import AIFLOW_HOME
from ai_flow.store.abstract_store import Filters, AbstractStore, Orders
from ai_flow.store.db.db_util import create_db_store
from ai_flow.util.json_utils import loads, Jsonable, dumps
from ai_flow.version import __version__
from ai_flow.workflow.control_edge import ControlEdge
from django.core.paginator import Paginator
from flask import Flask, request, render_template, Response, make_response, redirect
from flask_cors import CORS
from logging.config import dictConfig
from typing import List, Dict
from werkzeug.local import LocalProxy


def config_logging():
    dictConfig({
        'version': 1,
        'formatters': {'default': {
            'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        }},
        'handlers': {'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }},
        'root': {
            'level': 'INFO',
            'handlers': ['wsgi']
        }
    })


app = Flask(__name__, static_folder='./dist', template_folder='./dist', static_url_path='')
CORS(app=app)

store: AbstractStore = None
scheduler: Scheduler = None
airflow: str = None
config: AIFlowServerConfig = None
logger = logging.getLogger(__name__)


def init(store_uri: str, scheduler_class: str, airflow_web_server_uri: str):
    global store
    store = create_db_store(store_uri)
    global scheduler
    scheduler = SchedulerFactory.create_scheduler(scheduler_class,
                                                  {'notification_server_uri': None, 'airflow_deploy_path': None})
    global airflow
    airflow = airflow_web_server_uri


class Edge(Jsonable):
    def __init__(self, id: str = None, name: str = None, is_signal: int = 1, is_closed_loop_node: bool = True,
                 dag_data_type: str = None):
        self._id = id
        self._name = name
        self._is_signal = is_signal
        self._is_closed_loop_node = is_closed_loop_node
        self._dag_data_type = dag_data_type

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    def to_dict(self):
        return {'id': self._id, 'name': self._name, 'isSignal': self._is_signal,
                'isClosedLoopNode': self._is_closed_loop_node,
                'dagDataType': self._dag_data_type}


class Node(Jsonable):

    def __init__(self, id: str = None, layer: int = None, parent: List[Edge] = None, children: List[Edge] = None,
                 node_type: int = None, is_virtual: int = None, name: str = None, job_type_name=None,
                 is_real_node: bool = True, job_id: int = None, source_flag: bool = None, rsuuid: str = None,
                 data_name: str = None, source_type: str = None, material_id: str = None, material_ready: bool = False,
                 material_parent_project_id: int = None):
        self._id = id
        self._layer = layer
        self._parent = parent
        self._children = children
        self._node_type = node_type
        self._is_virtual = is_virtual
        self._name = name
        self._job_type_name = job_type_name
        self._is_real_node = is_real_node
        self._job_id = job_id
        self._source_flag = source_flag
        self._rsuuid = rsuuid
        self._data_name = data_name
        self._source_type = source_type
        self._material_id = material_id
        self._material_ready = material_ready
        self._material_parent_project_id = material_parent_project_id

    @property
    def id(self) -> str:
        return self._id

    @property
    def layer(self) -> int:
        return self._layer

    @layer.setter
    def layer(self, layer: int):
        self._layer = layer

    @property
    def parent(self) -> List[Edge]:
        return self._parent

    @parent.setter
    def parent(self, parent: List[Edge]):
        self._parent = parent

    @property
    def children(self) -> List[Edge]:
        return self._children

    @children.setter
    def children(self, children: List[Edge]):
        self._children = children

    @property
    def name(self) -> str:
        return self._name

    @property
    def data_name(self) -> str:
        return self._data_name

    def to_dict(self):
        return {'id': self._id, 'layer': self._layer,
                'parent': None if self._parent is None else [edge.to_dict() for edge in self._parent],
                'children': None if self._children is None else [edge.to_dict() for edge in self._children],
                'nodeType': self._node_type, 'isVirtual': self._is_virtual, 'name': self._name,
                'jobTypeName': self._job_type_name,
                'isRealNode': self._is_real_node, 'jobId': self._job_id, 'sourceFlag': self._source_flag,
                'rsuuid': self._rsuuid, 'dataName': self._data_name,
                'sourceType': self._source_type, 'materialId': self._material_id, 'materialReady': self._material_ready,
                'materialParentProjectId': self._material_parent_project_id}


def node_layer(node: Node, parent_edges: Dict, nodes: Dict, parent_nodes: List = None):
    if node.id not in parent_edges:
        return 1
    else:
        max_layer = 1
        parent_nodes = parent_nodes if parent_nodes else [node.id]
        for parent_edge in parent_edges[node.id]:
            if parent_edge.id not in parent_nodes:
                parent_nodes.append(parent_edge.id)
                layer = 1 + node_layer(nodes[parent_edge.id], parent_edges, nodes, parent_nodes)
                if layer > max_layer:
                    max_layer = layer
        return max_layer


def generate_graph(workflow_meta: WorkflowMeta):
    workflow_graph: AIGraph = extract_graph(workflow_meta)
    workflow_nodes, id_nodes, name_nodes = build_nodes(workflow_graph)
    parent_edges, children_edges = build_edges(workflow_graph, workflow_nodes, id_nodes, name_nodes)
    return build_graph(name_nodes, parent_edges, children_edges)


def extract_graph(workflow_meta: WorkflowMeta):
    graph_meta: Dict[str, str] = json.loads(workflow_meta.graph, strict=False)
    if '_context_extractor' in graph_meta:
        graph_meta.pop('_context_extractor', None)
    return loads(dumps(graph_meta))


def build_nodes(workflow_graph: AIGraph):
    workflow_nodes: Dict[str, AINode] = {}
    id_nodes: Dict[str, Node] = {}
    name_nodes: Dict[str, Node] = {}
    for graph_node in workflow_graph.nodes.values():
        workflow_nodes.update({graph_node.node_id: graph_node})
        if isinstance(graph_node, ReadDatasetNode) or isinstance(graph_node, WriteDatasetNode):
            data_node = Node(id=graph_node.dataset().name, source_flag=True,
                             data_name=graph_node.dataset().name)
            id_nodes.update({graph_node.node_id: data_node})
            name_nodes.update({graph_node.dataset().name: data_node})
        else:
            job_node = Node(id=graph_node.config.job_name, node_type=0, name=graph_node.config.job_name,
                            job_type_name=graph_node.config.job_type, source_flag=False)
            id_nodes.update({graph_node.node_id: job_node})
            name_nodes.update({graph_node.config.job_name: job_node})
    return workflow_nodes, id_nodes, name_nodes


def build_edges(workflow_graph: AIGraph, workflow_nodes: Dict[str, AINode], id_nodes: Dict[str, Node],
                name_nodes: Dict[str, Node]):
    parent_edges: Dict[str, List[Edge]] = {}
    children_edges: Dict[str, List[Edge]] = {}
    for graph_edges in workflow_graph.edges.values():
        for graph_edge in graph_edges:
            if isinstance(graph_edge, DataEdge):
                source_workflow_node: AINode = workflow_nodes.get(graph_edge.source)
                destination_workflow_node: AINode = workflow_nodes.get(graph_edge.destination)
                dag_data_type = 'source' if isinstance(source_workflow_node, ReadDatasetNode) else 'sink'
                source_node: Node = id_nodes.get(graph_edge.source)
                source_name = source_node.data_name \
                    if isinstance(source_workflow_node, ReadDatasetNode) \
                    else source_node.name
                source_edge: Edge = Edge(id=source_name, name=source_name, dag_data_type=dag_data_type, )
                destination_node: Node = id_nodes.get(graph_edge.destination)
                destination_name = destination_node.data_name \
                    if isinstance(destination_workflow_node, WriteDatasetNode) \
                    else destination_node.name
                destination_edge: Edge = Edge(id=destination_name, name=destination_name,
                                              dag_data_type=dag_data_type)
                if source_name != destination_name:
                    if source_name in children_edges:
                        children_edges[source_name].append(destination_edge)
                    else:
                        children_edges[source_name] = [destination_edge]
                    if destination_name in parent_edges:
                        parent_edges[destination_name].append(source_edge)
                    else:
                        parent_edges[destination_name] = [source_edge]
            else:
                control_edge: ControlEdge = graph_edge
                for event in control_edge.scheduling_rule.event_condition.events:
                    if event.sender != '*':
                        sender_event_edge: Edge = Edge(id=name_nodes[event.sender].id,
                                                       name=name_nodes[event.sender].name,
                                                       dag_data_type='event')
                        receiver_event_edge: Edge = Edge(id=name_nodes[control_edge.destination].id,
                                                         name=name_nodes[control_edge.destination].name,
                                                         dag_data_type='event')
                        if name_nodes[event.sender].id in children_edges:
                            children_edges[name_nodes[event.sender].id].append(receiver_event_edge)
                        else:
                            children_edges[name_nodes[event.sender].id] = [receiver_event_edge]
                        if name_nodes[control_edge.destination].id in parent_edges:
                            parent_edges[name_nodes[control_edge.destination].id].append(sender_event_edge)
                        else:
                            parent_edges[name_nodes[control_edge.destination].id] = [sender_event_edge]
    return parent_edges, children_edges


def build_graph(name_nodes: Dict[str, Node], parent_edges: Dict[str, List[Edge]],
                children_edges: Dict[str, List[Edge]]):
    node_layers = {}
    for name_node in name_nodes.values():
        name_node.layer = node_layer(name_node, parent_edges, name_nodes)
        if name_node.id in parent_edges:
            name_node.parent = parent_edges[name_node.id]
        if name_node.id in children_edges:
            name_node.children = children_edges[name_node.id]
        node_layers.update({name_node.id: name_node})
    graph_nodes = []
    for graph_node in node_layers.values():
        if graph_node.parent:
            for parent_node in graph_node.parent:
                node_layers[parent_node.id].layer = graph_node.layer - 1
        graph_nodes.append(graph_node.to_dict())
    return json.dumps(graph_nodes)


def store_inner_class(inner_name: str):
    return getattr(sys.modules[store.__module__], inner_name)


def build_filters(req: LocalProxy):
    filters = Filters()
    for key, value in req.args.items():
        if key not in ('pageNo', 'pageSize', 'sortField', 'sortOrder') and value:
            filters.add_filter((store_inner_class('FilterEqual')(key), value))
    return filters


def build_orders(req: LocalProxy):
    orders = Orders()
    if 'sortField' in req.args.keys() and 'sortOrder' in req.args.keys():
        orders.add_order((store_inner_class('OrderBy')(req.args['sortField']), req.args['sortOrder']))
    return orders


def pagination_response(page_no: int, total_count: int, data: List):
    return dumps({'pageNo': page_no, 'totalCount': total_count, 'data': data})


def json_pagination_response(page_no: int, total_count: int, data: List):
    return json.dumps({'pageNo': page_no, 'totalCount': total_count, 'data': data})


def make_zip(source_dir, target_file):
    if os.path.exists(target_file):
        os.remove(target_file)
    zip_file = zipfile.ZipFile(os.path.join(source_dir, target_file), 'w')
    for parent, dirs, files in os.walk(source_dir):
        for file in files:
            if file != target_file:
                file_path = os.path.join(parent, file)
                zip_file.write(file_path, file_path[len(os.path.dirname(source_dir)):].strip(os.path.sep))
    zip_file.close()


@app.route('/')
def index():
    return render_template('index.html')


@app.route('/metadata/<meta>')
def metadata(meta):
    return render_template('index.html')


@app.route('/graph/<project_id>/<workflow_name>')
def graph(project_id, workflow_name):
    return render_template('index.html')


@app.route('/version')
def version():
    return __version__


@app.route('/project')
def project_metadata():
    project_count = store.count_projects(filters=build_filters(request))
    project_list = store.list_projects(page_size=int(request.args.get('pageSize')),
                                       offset=(int(request.args.get('pageNo')) - 1) * int(request.args.get('pageSize')),
                                       filters=build_filters(request), orders=build_orders(request))
    return pagination_response(page_no=int(request.args.get('pageNo')), total_count=project_count,
                               data=project_list if project_list else [])


@app.route('/workflow')
def workflow_metadata():
    workflow_count = store.count_workflows(filters=build_filters(request))
    workflow_list = store.list_workflows(page_size=int(request.args.get('pageSize')),
                                         offset=(int(request.args.get('pageNo')) - 1) * int(
                                             request.args.get('pageSize')),
                                         filters=build_filters(request), orders=build_orders(request))
    return pagination_response(page_no=int(request.args.get('pageNo')), total_count=workflow_count,
                               data=workflow_list if workflow_list else [])


@app.route('/workflow/data-view')
def data_view():
    project_id = request.args.get('project_id')
    workflow_name = request.args.get('workflow_name')
    project_meta = store.get_project_by_id(project_id)
    if project_meta is None:
        raise Exception('The project({}) for the workflow({}) is not found.'.format(project_id, workflow_name))
    workflow_meta = store.get_workflow_by_name(project_name=project_meta.name, workflow_name=workflow_name)
    if workflow_meta is None:
        raise Exception('The workflow({}) of the project({}) is not found.'.format(workflow_name, project_id))
    else:
        return generate_graph(workflow_meta)


@app.route('/workflow/task-view')
def task_view():
    project_id = request.args.get('project_id')
    workflow_name = request.args.get('workflow_name')
    project_meta = store.get_project_by_id(project_id)
    if project_meta is None:
        raise Exception('The project({}) for the workflow({}) is not found.'.format(project_id, workflow_name))
    workflow_meta = store.get_workflow_by_name(project_name=project_meta.name, workflow_name=workflow_name)
    if workflow_meta is None:
        raise Exception('The workflow({}) of the project({}) is not found.'.format(workflow_name, project_id))
    else:
        return '{}/graph?dag_id={}'.format(airflow, '{}.{}'.format(project_meta.name, workflow_name))


@app.route('/workflow-execution')
def workflow_execution_metadata():
    project_name = request.args.get('project_name')
    workflow_name = request.args.get('workflow_name')
    workflow_execution_list = scheduler.list_workflow_executions(project_name,
                                                                 workflow_name) if project_name and workflow_name else None
    return pagination_response(page_no=int(request.args.get('pageNo')),
                               total_count=len(workflow_execution_list) if workflow_execution_list else 0,
                               data=Paginator(workflow_execution_list, int(request.args.get('pageSize'))).get_page(
                                   int(request.args.get('pageNo'))).object_list if workflow_execution_list else [])


@app.route('/job-execution')
def job_execution_metadata():
    workflow_execution_id = request.args.get('workflow_execution_id')
    job_execution_list = scheduler.list_job_executions(workflow_execution_id) if workflow_execution_id else None
    page_job_executions = Paginator(job_execution_list, int(request.args.get('pageSize'))).get_page(
      int(request.args.get('pageNo'))).object_list if job_execution_list else None
    job_execution_objects = []
    if page_job_executions:
        workflow_info = page_job_executions[0].workflow_execution.workflow_info
        workflow_graph = extract_graph(store.get_workflow_by_name(workflow_info.namespace,
                                                                  workflow_info.workflow_name))
        job_types = {}
        for node in workflow_graph.nodes.values():
            job_types.update({node.config.job_name: node.config.job_type})
        for page_job_execution in page_job_executions:
            job_execution_object = page_job_execution.__dict__
            job_execution_object.update({'_job_type': job_types.get(page_job_execution.job_name)})
            job_execution_objects.append(job_execution_object)
    return pagination_response(page_no=int(request.args.get('pageNo')),
                                total_count=len(job_execution_list) if job_execution_list else 0,
                                data=job_execution_objects)


@app.route('/job-execution/log')
def job_execution_log():
    job_type = request.args.get('job_type')
    job_execution_id = request.args.get('job_execution_id')
    workflow_execution_id = request.args.get('workflow_execution_id')
    job_executions = scheduler.get_job_executions(request.args.get('job_name'),
                                                  workflow_execution_id)
    log_job_execution = None
    for job_execution in job_executions:
        if job_execution.job_execution_id == job_execution_id:
            log_job_execution = job_execution
            break
    if log_job_execution:
        if job_type == 'bash':
            return redirect('{}/graph?dag_id={}'.format(airflow, workflow_execution_id.split('|')[0]))
        else:
            base_log_folder = config.get_base_log_folder()
            log_dir = os.path.join(base_log_folder if base_log_folder else AIFLOW_HOME, 'logs',
                                   log_job_execution.workflow_execution.workflow_info.namespace,
                                   log_job_execution.workflow_execution.workflow_info.workflow_name,
                                   log_job_execution.job_name,
                                   log_job_execution.job_execution_id)
            if os.path.exists(log_dir):
                log_file = 'logs.zip'
                make_zip(log_dir, log_file)
                file_obj = io.BytesIO()
                with zipfile.ZipFile(file_obj, 'w') as zip_file:
                    zip_info = zipfile.ZipInfo(log_file)
                    zip_info.date_time = time.localtime(time.time())[:6]
                    zip_info.compress_type = zipfile.ZIP_DEFLATED
                    with open(os.path.join(log_dir, log_file), 'rb') as fd:
                        zip_file.writestr(zip_info, fd.read())
                file_obj.seek(0)
                return Response(file_obj.getvalue(),
                                mimetype='application/zip',
                                headers={'Content-Disposition': 'attachment;filename={}'.format(log_file)})
    return make_response('No log found.')


@app.route('/dataset')
def dataset_metadata():
    dataset_count = store.count_datasets(filters=build_filters(request))
    dataset_list = store.list_datasets(page_size=int(request.args.get('pageSize')),
                                       offset=(int(request.args.get('pageNo')) - 1) * int(request.args.get('pageSize')),
                                       filters=build_filters(request), orders=build_orders(request))
    return pagination_response(page_no=int(request.args.get('pageNo')), total_count=dataset_count,
                               data=dataset_list if dataset_list else [])


@app.route('/model')
def model_metadata():
    model_count = store.count_registered_models(filters=build_filters(request))
    model_list = store.list_registered_models(page_size=int(request.args.get('pageSize')),
                                              offset=(int(request.args.get('pageNo')) - 1) * int(
                                                  request.args.get('pageSize')),
                                              filters=build_filters(request), orders=build_orders(request))
    return json_pagination_response(page_no=int(request.args.get('pageNo')), total_count=model_count,
                                    data=[{'model_name': model.model_name, 'model_desc': model.model_desc} for model in
                                          model_list] if model_list else [])


@app.route('/model-version')
def model_version_metadata():
    model_version_count = store.count_model_versions(filters=build_filters(request))
    model_version_list = store.list_model_versions(page_size=int(request.args.get('pageSize')),
                                                   offset=(int(request.args.get('pageNo')) - 1) * int(
                                                       request.args.get('pageSize')),
                                                   filters=build_filters(request), orders=build_orders(request))
    return json_pagination_response(page_no=int(request.args.get('pageNo')), total_count=model_version_count,
                                    data=[model_version.__dict__ for model_version in
                                          model_version_list] if model_version_list else [])


@app.route('/artifact')
def artifact_metadata():
    artifact_count = store.count_artifacts(filters=build_filters(request))
    artifact_list = store.list_artifacts(page_size=int(request.args.get('pageSize')),
                                         offset=(int(request.args.get('pageNo')) - 1) * int(
                                             request.args.get('pageSize')),
                                         filters=build_filters(request), orders=build_orders(request))
    return pagination_response(page_no=int(request.args.get('pageNo')), total_count=artifact_count,
                               data=artifact_list if artifact_list else [])


usage_message = """
usage: web_server.py -c <config_path>
"""


class AIFlowWebServerConfig:

    def __init__(self, config: Dict):
        self.config = config

    def host(self):
        return self.config['host']

    def port(self):
        return self.config['port']

    def get(self, key):
        return self.config.get(key)

    def __repr__(self):
        return self.config.__repr__()

    def __eq__(self, other):
        return self.config.__eq__(other.config)


def main(argv):
    config_logging()
    config_file = None
    try:
        opts, args = getopt.getopt(argv, "hc:",
                                   ["config="])
    except getopt.GetoptError:
        print(usage_message)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print(usage_message)
            sys.exit()
        elif opt in ("-c", "--config"):
            config_file = arg

    if config_file is None:
        print(usage_message)
        sys.exit(2)

    start_web_server_by_config_file_path(config_file)


def start_web_server_by_config_file_path(config_file_path):
    global config
    config = AIFlowServerConfig()
    config.load_from_file(config_file_path)
    store_uri = config.get_db_uri()
    scheduler_service_config = SchedulerServiceConfig(config.get_scheduler_service_config())
    scheduler_class = scheduler_service_config.scheduler().scheduler_class()
    web_server_config = AIFlowWebServerConfig(config.get_web_server_config())

    start_web_server(scheduler_class, store_uri, web_server_config)


def start_web_server(scheduler_class, store_uri, web_server_config):
    logger.info("Starting web server with scheduler class: {}, store_uri: {}, web_server_config: {}"
                .format(scheduler_class, store_uri, web_server_config))
    init(store_uri, scheduler_class, web_server_config.get('airflow_web_server_uri'))
    app.run(host=web_server_config.host(), port=int(web_server_config.port()))


if __name__ == '__main__':
    main(sys.argv[1:])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import inspect
import os
import sys
import tempfile
import traceback
from typing import Text

from ai_flow.client.ai_flow_client import get_ai_flow_client
from ai_flow.context.project_context import init_project_context, current_project_config, set_current_project_config
from ai_flow.context.workflow_config_loader import init_workflow_config

__init_context_flag__ = False
__init_notebook_context_flag__ = False
__init_client_flag__ = False

from ai_flow.project.project_config import ProjectConfig
from ai_flow.workflow.workflow_config import WorkflowConfig, dump_workflow_config


def init_ai_flow_context():
    """
    When defining the ai flow program,
    you need to call this function to initialize the project and workflow environment information.
    This function has three functions:
    1. Init project context
    2. Init project configuration
    3. Init workflow configuration.
    """
    if __init_notebook_context_flag__:
        raise Exception('init_ai_flow_context and init_notebook_context cannot be called at the same time.')
    if __init_client_flag__:
        raise Exception('init_ai_flow_client and init_ai_flow_context cannot be called at the same time.')
    stack = traceback.extract_stack()
    workflow_entry_file = os.path.abspath(stack[-2].filename)
    workflows_path = os.path.dirname(os.path.dirname(workflow_entry_file))
    # workflow_name/workflow_name.py len(.py) == 3
    workflow_name = os.path.basename(workflow_entry_file)[:-3]
    project_path = os.path.dirname(workflows_path)
    init_project_context(project_path)
    ensure_project_registered()
    # workflow_name/workflow_name.yaml
    init_workflow_config(workflow_config_file
                         =os.path.join(workflows_path, workflow_name, '{}.yaml'.format(workflow_name)))
    global __init_context_flag__
    __init_context_flag__ = True


def init_notebook_context(project_config: ProjectConfig, workflow_config: WorkflowConfig):
    """
    Initializes the notebook context for the project and workflow environment when user defines the workflow:
    1. Generates project and workflow.
    2. Initializes project context.
    3. Initializes project configuration.
    4. Initializes workflow configuration.
    """
    if __init_context_flag__:
        raise Exception('init_notebook_context and init_ai_flow_context cannot be called at the same time.')

    def generate(workflow_code):
        """
        Generates the project and workflow automatically.

        :param workflow_code: User-defined workflow code.
        :return: Project path and workflow configuration path.
        """
        temp_path = tempfile.mkdtemp()
        temp_project_path = '{}/{}'.format(temp_path, project_config.get_project_name())
        workflow_name = workflow_config.workflow_name
        temp_workflow_path = '{}/workflows/{}'.format(temp_project_path, workflow_name)
        os.makedirs(temp_workflow_path)
        project_config.dump_to_file('{}/project.yaml'.format(temp_project_path))
        temp_workflow_config = '{}/{}.yaml'.format(temp_workflow_path, workflow_name)
        dump_workflow_config(workflow_config, temp_workflow_config)
        with open('{}/{}.py'.format(temp_workflow_path, workflow_name), 'w') as f:
            f.write(workflow_code)
        return temp_project_path, temp_workflow_config

    project_path, workflow_config_path = generate(inspect.getsource(sys._getframe().f_back.f_back))
    init_project_context(project_path)
    ensure_project_registered()
    init_workflow_config(workflow_config_path)
    global __init_notebook_context_flag__
    __init_notebook_context_flag__ = True


def ensure_project_registered():
    """ Ensure the project configured in project.yaml has been registered. """

    client = get_ai_flow_client()
    project_meta = client.get_project_by_name(current_project_config().get_project_name())
    pp = {}
    for k, v in current_project_config().items():
        pp[k] = str(v)
    if project_meta is None:
        project_meta = client.register_project(name=current_project_config().get_project_name(),
                                               properties=pp)
    else:
        project_meta = client.update_project(project_name=current_project_config().get_project_name(), properties=pp)

    current_project_config().set_project_uuid(str(project_meta.uuid))


def init_ai_flow_client(server_uri: Text, project_name: Text = None, **kwargs):
    """ Initializes the :class:`ai_flow.client.ai_flow_client.AIFlowClient`.
        It's suitable for using AIFlowClient separately in per job.
    """
    if __init_context_flag__ is True:
        raise Exception('init_ai_flow_client and init_ai_flow_context cannot be called at the same time.')
    config_dict = {'server_uri': server_uri}
    if project_name is None:
        project_name = 'Unknown'
    config_dict['project_name'] = project_name
    for k, v in kwargs.items():
        config_dict[k] = v
    set_current_project_config(config_dict)
    ensure_project_registered()
    global __init_client_flag__
    __init_client_flag__ = True



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""task_instance_add_is_schedulable

Revision ID: 818f168686ac
Revises: cf3cb8473f3c
Create Date: 2021-12-28 15:26:13.728326

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = '818f168686ac'
down_revision = 'cf3cb8473f3c'
branch_labels = None
depends_on = None


def upgrade():
    """Apply task_instance_add_is_schedulable"""
    op.add_column('task_instance', sa.Column('is_schedulable', sa.Boolean, nullable=True, default=True))


def downgrade():
    """Unapply task_instance_add_is_schedulable"""
    op.drop_column('task_instance', 'is_schedulable')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest
from notification_service.base_notification import BaseEvent
from ai_flow.client.notification_client import get_notification_client
from ai_flow.context.project_context import set_current_project_config
from ai_flow.context.job_context import set_current_job_name, unset_current_job_name
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server, _NS_URI


class TestNotificationClient(unittest.TestCase):

    def setUp(self) -> None:
        self.ns_server = start_notification_server()

    def tearDown(self) -> None:
        stop_notification_server(self.ns_server)

    def test_get_notification_client(self):
        config = {'notification_server_uri': _NS_URI, 'project_name': 'test_project'}
        set_current_project_config(config)
        set_current_job_name('job_1')
        client = get_notification_client()
        client.send_event(BaseEvent(key='a', value='a'))
        events = client.list_all_events(start_version=0)
        self.assertEqual(1, len(events))
        self.assertEqual('test_project', events[0].namespace)
        self.assertEqual('job_1', events[0].sender)
        self.assertEqual('a', events[0].key)
        self.assertEqual('a', events[0].value)
        unset_current_job_name()


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import cloudpickle

from ai_flow.api.context_extractor import ContextExtractor

from ai_flow.util.json_utils import Jsonable
from typing import Text, List
from ai_flow.common.properties import Properties
from ai_flow.workflow.control_edge import WorkflowSchedulingRule, WorkflowAction, EventCondition


class WorkflowMeta(Jsonable):
    """define workflow meta"""

    def __init__(self,
                 name: Text,
                 project_id: int,
                 properties: Properties = None,
                 create_time: int = None,
                 update_time: int = None,
                 context_extractor_in_bytes: bytes = None,
                 graph: str = None,
                 uuid: int = None,
                 scheduling_rules: List[WorkflowSchedulingRule] = None,
                 last_event_version: int = None
                 ) -> None:
        """

        :param name: the workflow name
        :param project_id: the uuid of project which contains this workflow
        :param properties: the workflow properties
        :param create_time: create time represented as milliseconds since epoch.
        :param update_time: update time represented as milliseconds since epoch.
        :param context_extractor_in_bytes: the user-defined logic of how to extract context from event
        :param graph: the graph of the workflow
        :param uuid: uuid in database
        :param scheduling_rules: A list of scheduling rules of the workflow.
        :param last_event_version: last processed event version of the workflow. It is none, if there are no event
        processed after submit.
        """

        self.name = name
        self.project_id = project_id
        self.properties = properties
        self.create_time = create_time
        self.update_time = update_time
        self.uuid = uuid
        self.context_extractor_in_bytes = context_extractor_in_bytes
        self.graph = graph
        self.scheduling_rules = [] if scheduling_rules is None else scheduling_rules
        self.last_event_version = last_event_version

    def get_condition(self, action: WorkflowAction) -> List[EventCondition]:
        if self.scheduling_rules is None:
            return []
        return [rule.event_condition for rule in self.scheduling_rules if rule.action == action]

    def update_condition(self, event_conditions: List[EventCondition], action: WorkflowAction):
        scheduling_rules = self.scheduling_rules if self.scheduling_rules is not None else []
        new_scheduling_rules = [rule for rule in scheduling_rules if rule.action != action]
        new_scheduling_rules.extend(
            [WorkflowSchedulingRule(condition, action) for condition in event_conditions])
        self.scheduling_rules = new_scheduling_rules

    def __str__(self):
        return '<\n' \
               'WorkflowMeta\n' \
               'uuid:{},\n' \
               'name:{},\n' \
               'project_id:{},\n' \
               'properties:{},\n' \
               'create_time:{},\n' \
               'update_time:{},\n' \
               'context_extractor_in_bytes:{},\n' \
               'graph:{},\n' \
               '>'.format(self.uuid,
                          self.name,
                          self.project_id,
                          self.properties,
                          self.create_time,
                          self.update_time,
                          self.context_extractor_in_bytes,
                          self.graph)

    def get_context_extractor(self) -> ContextExtractor:
        """
        Return the deserialized ContextExtractor instance.

        """
        return cloudpickle.loads(self.context_extractor_in_bytes)

    def set_context_extractor(self, context_extractor: ContextExtractor):
        """
        Set the context_extractor_in_bytes from given ContextExtractor instance.

        :param context_extractor: ContextExtractor instance
        """
        self.context_extractor_in_bytes = cloudpickle.dumps(context_extractor)


def create_workflow(name: Text,
                    project_id: int,
                    properties: Properties = None,
                    create_time: int = None,
                    update_time: int = None,
                    context_extractor_in_bytes: bytes = None,
                    graph: str = None
                    ) -> WorkflowMeta:
    return WorkflowMeta(name=name, project_id=int(project_id), properties=properties,
                        create_time=create_time, update_time=update_time,
                        context_extractor_in_bytes=context_extractor_in_bytes,
                        graph=graph)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""Utilities module for cli"""
import contextlib
import functools
import getpass
import json
import logging
import os
import re
import socket
import sys
import threading
import traceback
import warnings
from argparse import Namespace
from datetime import datetime
from typing import TYPE_CHECKING, Callable, Optional, TypeVar, cast

from airflow import settings
from airflow.exceptions import AirflowException
from airflow.utils import cli_action_loggers
from airflow.utils.platform import is_terminal_support_colors
from airflow.utils.session import provide_session
from airflow.utils.file import get_sha1hash

T = TypeVar("T", bound=Callable)  # pylint: disable=invalid-name

if TYPE_CHECKING:
    from airflow.models import DAG


def action_logging(f: T) -> T:
    """
    Decorates function to execute function at the same time submitting action_logging
    but in CLI context. It will call action logger callbacks twice,
    one for pre-execution and the other one for post-execution.

    Action logger will be called with below keyword parameters:
        sub_command : name of sub-command
        start_datetime : start datetime instance by utc
        end_datetime : end datetime instance by utc
        full_command : full command line arguments
        user : current user
        log : airflow.models.log.Log ORM instance
        dag_id : dag id (optional)
        task_id : task_id (optional)
        execution_date : execution date (optional)
        error : exception instance if there's an exception

    :param f: function instance
    :return: wrapped function
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        """
        An wrapper for cli functions. It assumes to have Namespace instance
        at 1st positional argument

        :param args: Positional argument. It assumes to have Namespace instance
            at 1st positional argument
        :param kwargs: A passthrough keyword argument
        """
        if not args:
            raise ValueError("Args should be set")
        if not isinstance(args[0], Namespace):
            raise ValueError(
                "1st positional argument should be argparse.Namespace instance," f"but is {type(args[0])}"
            )
        metrics = _build_metrics(f.__name__, args[0])
        cli_action_loggers.on_pre_execution(**metrics)
        try:
            return f(*args, **kwargs)
        except Exception as e:
            metrics['error'] = e
            raise
        finally:
            metrics['end_datetime'] = datetime.utcnow()
            cli_action_loggers.on_post_execution(**metrics)

    return cast(T, wrapper)


def _build_metrics(func_name, namespace):
    """
    Builds metrics dict from function args
    It assumes that function arguments is from airflow.bin.cli module's function
    and has Namespace instance where it optionally contains "dag_id", "task_id",
    and "execution_date".

    :param func_name: name of function
    :param namespace: Namespace instance from argparse
    :return: dict with metrics
    """
    from airflow.models import Log

    sensitive_fields = {'-p', '--password', '--conn-password'}
    full_command = list(sys.argv)
    for idx, command in enumerate(full_command):  # pylint: disable=too-many-nested-blocks
        if command in sensitive_fields:
            # For cases when password is passed as "--password xyz" (with space between key and value)
            full_command[idx + 1] = "*" * 8
        else:
            # For cases when password is passed as "--password=xyz" (with '=' between key and value)
            for sensitive_field in sensitive_fields:
                if command.startswith(f'{sensitive_field}='):
                    full_command[idx] = f'{sensitive_field}={"*" * 8}'

    metrics = {
        'sub_command': func_name,
        'start_datetime': datetime.utcnow(),
        'full_command': f'{full_command}',
        'user': getpass.getuser(),
    }

    if not isinstance(namespace, Namespace):
        raise ValueError(
            "namespace argument should be argparse.Namespace instance," f"but is {type(namespace)}"
        )
    tmp_dic = vars(namespace)
    metrics['dag_id'] = tmp_dic.get('dag_id')
    metrics['task_id'] = tmp_dic.get('task_id')
    metrics['execution_date'] = tmp_dic.get('execution_date')
    metrics['host_name'] = socket.gethostname()

    extra = json.dumps({k: metrics[k] for k in ('host_name', 'full_command')})
    log = Log(
        event=f'cli_{func_name}',
        task_instance=None,
        owner=metrics['user'],
        extra=extra,
        task_id=metrics.get('task_id'),
        dag_id=metrics.get('dag_id'),
        execution_date=metrics.get('execution_date'),
    )
    metrics['log'] = log
    return metrics


def process_subdir(subdir: Optional[str]):
    """Expands path to absolute by replacing 'DAGS_FOLDER', '~', '.', etc."""
    if subdir:
        if not settings.DAGS_FOLDER:
            raise ValueError("DAGS_FOLDER variable in settings should be filled.")
        subdir = subdir.replace('DAGS_FOLDER', settings.DAGS_FOLDER)
        subdir = os.path.abspath(os.path.expanduser(subdir))
    return subdir


def get_dag_by_file_location(dag_id: str):
    """Returns DAG of a given dag_id by looking up file location"""
    from airflow.models import DagBag, DagModel

    # Benefit is that logging from other dags in dagbag will not appear
    dag_model = DagModel.get_current(dag_id)
    if dag_model is None:
        raise AirflowException(
            'dag_id could not be found: {}. Either the dag did not exist or it failed to '
            'parse.'.format(dag_id)
        )
    dagbag = DagBag(dag_folder=dag_model.fileloc)
    return dagbag.dags[dag_id]


def get_dag(subdir: Optional[str], dag_id: str) -> "DAG":
    """Returns DAG of a given dag_id"""
    from airflow.models import DagBag

    dagbag = DagBag(process_subdir(subdir))
    if dag_id not in dagbag.dags:
        raise AirflowException(
            'dag_id could not be found: {}. Either the dag did not exist or it failed to '
            'parse.'.format(dag_id)
        )
    return dagbag.dags[dag_id]


def get_dags(subdir: Optional[str], dag_id: str, use_regex: bool = False):
    """Returns DAG(s) matching a given regex or dag_id"""
    from airflow.models import DagBag

    if not use_regex:
        return [get_dag(subdir, dag_id)]
    dagbag = DagBag(process_subdir(subdir))
    matched_dags = [dag for dag in dagbag.dags.values() if re.search(dag_id, dag.dag_id)]
    if not matched_dags:
        raise AirflowException(
            'dag_id could not be found with regex: {}. Either the dag did not exist '
            'or it failed to parse.'.format(dag_id)
        )
    return matched_dags


@provide_session
def get_dag_by_pickle(pickle_id, session=None):
    """Fetch DAG from the database using pickling"""
    from airflow.models import DagPickle

    dag_pickle = session.query(DagPickle).filter(DagPickle.id == pickle_id).first()
    if not dag_pickle:
        raise AirflowException("Who hid the pickle!? [missing pickle]")
    pickle_dag = dag_pickle.pickle
    return pickle_dag


def load_dag_from_db_if_needed(dag_file_path: str, expected_sha1hash: str):
    """
    If file does not exists or the sha1 hash of file is not same as expected,
    load dag file from database.

    param: dag_file_path: the dag file needs to be checked and loaded
    param: expected_sha1hash: the expected sha1 hash of dag file
    """
    from airflow.models.dagcode import DagCode

    if os.path.exists(dag_file_path):
        local_dag_file_hash = get_sha1hash(dag_file_path)
        if local_dag_file_hash == expected_sha1hash:
            return
    parent = os.path.dirname(dag_file_path)
    if not os.path.exists(parent):
        os.makedirs(parent)
    with open(dag_file_path, 'w') as dag_file:
        dag_file.write(DagCode.code(dag_file_path))


def setup_locations(process, pid=None, stdout=None, stderr=None, log=None):
    """Creates logging paths"""
    if not stderr:
        stderr = os.path.join(settings.AIRFLOW_HOME, f'airflow-{process}.err')
    if not stdout:
        stdout = os.path.join(settings.AIRFLOW_HOME, f'airflow-{process}.out')
    if not log:
        log = os.path.join(settings.AIRFLOW_HOME, f'airflow-{process}.log')
    if not pid:
        pid = os.path.join(settings.AIRFLOW_HOME, f'airflow-{process}.pid')

    return pid, stdout, stderr, log


def setup_logging(filename):
    """Creates log file handler for daemon process"""
    root = logging.getLogger()
    handler = logging.FileHandler(filename)
    formatter = logging.Formatter(settings.SIMPLE_LOG_FORMAT)
    handler.setFormatter(formatter)
    root.addHandler(handler)
    root.setLevel(settings.LOGGING_LEVEL)

    return handler.stream


def sigint_handler(sig, frame):  # pylint: disable=unused-argument
    """
    Returns without error on SIGINT or SIGTERM signals in interactive command mode
    e.g. CTRL+C or kill <PID>
    """
    sys.exit(0)


def sigquit_handler(sig, frame):  # pylint: disable=unused-argument
    """
    Helps debug deadlocks by printing stacktraces when this gets a SIGQUIT
    e.g. kill -s QUIT <PID> or CTRL+\
    """
    print(f"Dumping stack traces for all threads in PID {os.getpid()}")
    id_to_name = {th.ident: th.name for th in threading.enumerate()}
    code = []
    for thread_id, stack in sys._current_frames().items():  # pylint: disable=protected-access
        code.append("\n# Thread: {}({})".format(id_to_name.get(thread_id, ""), thread_id))
        for filename, line_number, name, line in traceback.extract_stack(stack):
            code.append(f'File: "{filename}", line {line_number}, in {name}')
            if line:
                code.append(f"  {line.strip()}")
    print("\n".join(code))


class ColorMode:
    """Coloring modes. If `auto` is then automatically detected."""

    ON = "on"
    OFF = "off"
    AUTO = "auto"


def should_use_colors(args) -> bool:
    """Processes arguments and decides whether to enable color in output"""
    if args.color == ColorMode.ON:
        return True
    if args.color == ColorMode.OFF:
        return False
    return is_terminal_support_colors()


@contextlib.contextmanager
def suppress_logs_and_warning():
    """Context manager to suppress logging and warning messages"""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        logging.disable(logging.CRITICAL)
        yield
        # logging output again depends on the effective
        # levels of individual loggers
        logging.disable(logging.NOTSET)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import shutil
import unittest

from ai_flow.ai_graph.ai_graph import current_graph
from ai_flow.cli import cli_parser
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.plugin_interface import job_plugin_interface, register_job_plugin_factory
from ai_flow.scheduler_service.service.config import SchedulerServiceConfig
from ai_flow.test.api.mock_plugins import MockJobFactory
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server
from ai_flow.test.util.server_util import wait_for_server_started
from ai_flow.util.path_util import get_file_dir, get_parent_dir

_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)
_PORT = '50051'
_SERVER_URI = 'localhost:{}'.format(_PORT)
_SCHEDULER_CLASS = 'ai_flow.test.api.mock_plugins.MockScheduler'
_TEMP_DIR = '/tmp/ai_flow_mock_blob_directory'

PROJECT_PATH = os.path.join(get_file_dir(__file__), 'ut_workflows')
WORKFLOW_NAME = 'test_command'


class TestCommand(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.parser = cli_parser.get_parser()
        cls.notification_server = start_notification_server()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        raw_config = {
            'scheduler': {
                'scheduler_class': _SCHEDULER_CLASS,
            }
        }
        config = SchedulerServiceConfig(raw_config)
        register_job_plugin_factory(MockJobFactory())
        cls.aiflow_server = AIFlowServer(store_uri=_SQLITE_DB_URI, port=_PORT,
                                         start_meta_service=True,
                                         start_metric_service=False,
                                         start_model_center_service=False,
                                         start_scheduler_service=True,
                                         scheduler_service_config=config)
        cls.aiflow_server.run()
        wait_for_server_started(_SERVER_URI)
        if not os.path.exists(_TEMP_DIR):
            os.makedirs(_TEMP_DIR)

    @classmethod
    def tearDownClass(cls):
        cls.aiflow_server.stop()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        stop_notification_server(cls.notification_server)
        job_plugin_interface.__job_controller_manager__.object_dict.pop('mock')
        if os.path.exists(_TEMP_DIR):
            shutil.rmtree(_TEMP_DIR)

    def tearDown(self):
        current_graph().clear_graph()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os
import logging
import sqlalchemy


_logger = logging.getLogger(__name__)


def _get_alembic_config(url):
    from alembic.config import Config

    current_dir = os.path.dirname(os.path.abspath(__file__))
    package_dir = os.path.normpath(os.path.join(current_dir, '..'))
    directory = os.path.join(package_dir, 'migrations')
    config = Config(os.path.join(package_dir, 'alembic.ini'))
    config.set_main_option('script_location', directory.replace('%', '%%'))
    config.set_main_option('sqlalchemy.url', url.replace('%', '%%'))
    return config


def upgrade(url, version=None):
    """Upgrade the database."""
    # alembic adds significant import time, so we import it lazily
    from alembic import command

    _logger.info("Upgrade the database, db uri: {}.".format(url))
    config = _get_alembic_config(url)
    if version is None:
        version = 'heads'
    command.upgrade(config, version)


def init_db(url):
    _logger.info("Init the database, db uri: {}.".format(url))
    upgrade(url)


def clear_db(url, metadata):
    _logger.info("Clear the database, db uri: {}.".format(url))
    engine = sqlalchemy.create_engine(url)
    connection = engine.connect()
    metadata.drop_all(connection)
    # alembic adds significant import time, so we import it lazily
    from alembic.migration import MigrationContext  # noqa

    migration_ctx = MigrationContext.configure(connection)
    version = migration_ctx._version  # noqa pylint: disable=protected-access
    if version.exists(connection):
        version.drop(connection)


def reset_db(url, metadata):
    _logger.info("Reset the database, db uri: {}.".format(url))
    clear_db(url, metadata)
    init_db(url)


def downgrade(url, version):
    """Downgrade the database."""
    # alembic adds significant import time, so we import it lazily
    from alembic import command

    _logger.info("Downgrade the database, db uri: {}.".format(url))
    config = _get_alembic_config(url)
    command.downgrade(config, version)


def tables_exists(url):
    """Returns whether tables are created"""
    engine_ = sqlalchemy.create_engine(url)
    if len(engine_.table_names()) > 0:
        return True
    else:
        return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""model_version_add_createtime

Revision ID: 3ba5526346e3
Revises: 0e8fefcce376
Create Date: 2022-02-15 17:17:04.220562

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3ba5526346e3'
down_revision = '0e8fefcce376'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('model_version') as batch_op:
        batch_op.alter_column('model_version', type_=sa.Integer())
        batch_op.add_column(sa.Column('create_time', sa.BigInteger(), nullable=True))
    with op.batch_alter_table('model_version_relation') as batch_op:
        batch_op.alter_column('version', type_=sa.Integer())
    with op.batch_alter_table('metric_summary') as batch_op:
        batch_op.alter_column('model_version', type_=sa.Integer())


def downgrade():
    with op.batch_alter_table('model_version') as batch_op:
        batch_op.alter_column('model_version', type_=sa.String(10))
        batch_op.drop_column('create_time')
    with op.batch_alter_table('model_version_relation') as batch_op:
        batch_op.alter_column('version', type_=sa.String(10))
    with op.batch_alter_table('metric_summary') as batch_op:
        batch_op.alter_column('model_version', type_=sa.String(10))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import queue


class MessageQueue(object):

    def __init__(self) -> None:
        self.queue = queue.Queue()

    def send(self, event):
        self.queue.put_nowait(event)

    def get(self):
        return self.queue.get()

    def length(self):
        return self.queue.qsize()

    def get_with_timeout(self, timeout=1):
        try:
            return self.queue.get(timeout=timeout)
        except Exception as e:
            return None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# 'License'); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Job command"""

from ai_flow.api.workflow_operation import list_job_executions, restart_job_execution, get_job_executions, \
    start_job_execution, stop_job_execution, stop_scheduling_job, resume_scheduling_job
from ai_flow.cli.simple_table import AIFlowConsole
from ai_flow.util.cli_utils import init_config
from ai_flow.util.time_utils import parse_date


@init_config
def job_list_executions(args):
    """Lists all job executions of the workflow execution by workflow execution id."""
    job_executions = list_job_executions(args.workflow_execution_id)
    AIFlowConsole().print_as(
        data=sorted(job_executions, key=lambda j: j.job_execution_id),
        output=args.output,
        mapper=lambda x: {
            'job_execution_id': x.job_execution_id,
            'job_name': x.job_name,
            'workflow_execution_id': x.workflow_execution.workflow_execution_id,
            'status': x.status,
            'properties': x.properties,
            'start_date': parse_date(x.start_date),
            'end_date': parse_date(x.end_date),
            'execution_label': x.execution_label
        },
    )


@init_config
def job_restart_execution(args):
    """Restarts the job execution by job name and workflow execution id."""
    job_execution = restart_job_execution(args.job_name, args.workflow_execution_id)
    print("Job: {}, workflow execution: {}, restarted: {}.".format(args.job_name, args.workflow_execution_id,
                                                                   job_execution is not None))


@init_config
def job_show_execution(args):
    """Shows the job execution by job name and workflow execution id."""
    job_executions = get_job_executions(args.job_name, args.workflow_execution_id)
    AIFlowConsole().print_as(
        data=sorted(job_executions, key=lambda j: j.job_execution_id),
        output=args.output,
        mapper=lambda x: {
            'job_execution_id': x.job_execution_id,
            'job_name': x.job_name,
            'workflow_execution_id': x.workflow_execution.workflow_execution_id,
            'status': x.status,
            'properties': x.properties,
            'start_date': parse_date(x.start_date),
            'end_date': parse_date(x.end_date),
            'execution_label': x.execution_label
        },
    )


@init_config
def job_start_execution(args):
    """Starts the job execution by job name and workflow execution id."""
    job_execution = start_job_execution(args.job_name, args.workflow_execution_id)
    print("Job: {}, workflow execution: {}, started: {}.".format(args.job_name, args.workflow_execution_id,
                                                                 job_execution is not None))


@init_config
def job_stop_execution(args):
    """Stops the job execution by job name and workflow execution id."""
    job_execution = stop_job_execution(args.job_name, args.workflow_execution_id)
    print("Job: {}, workflow execution: {}, stopped: {}.".format(args.job_name, args.workflow_execution_id,
                                                                 job_execution is not None))


@init_config
def job_stop_scheduling(args):
    """Stops scheduling the job by job name and workflow execution id."""
    stop_scheduling_job(job_name=args.job_name, execution_id=args.workflow_execution_id)
    print("Job: {}, workflow execution: {}, stop scheduling.".format(args.job_name, args.workflow_execution_id))


@init_config
def job_resume_scheduling(args):
    """Resumes scheduling the job by job name and workflow execution id."""
    resume_scheduling_job(job_name=args.job_name, execution_id=args.workflow_execution_id)
    print("Job: {}, workflow execution: {}, resume scheduling.".format(args.job_name, args.workflow_execution_id))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add seq_num to task_instance

Revision ID: c12f471f01f4
Revises: d47076ca1426
Create Date: 2021-09-08 14:36:44.693163

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
from sqlalchemy.engine.reflection import Inspector
# revision identifiers, used by Alembic.
revision = 'c12f471f01f4'
down_revision = 'd47076ca1426'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()  # pylint: disable=no-member
    is_sqlite = bool(conn.dialect.name == "sqlite")

    if is_sqlite:
        op.execute("PRAGMA foreign_keys=off")

    with op.batch_alter_table('task_instance', schema=None) as batch_op:
        batch_op.add_column(sa.Column('is_active', sa.Boolean(), nullable=True))
        batch_op.add_column(sa.Column('seq_num', sa.Integer(), nullable=True))

    with op.batch_alter_table('task_execution', schema=None) as batch_op:
        batch_op.add_column(sa.Column('try_number', sa.Integer(), nullable=True))


def downgrade():
    """Unapply add seq_num to task_instance"""
    with op.batch_alter_table('task_instance', schema=None) as batch_op:
        batch_op.drop_column('is_active')
        batch_op.drop_column('seq_num')

    """Unapply add try_number to task_execution"""
    with op.batch_alter_table('task_execution', schema=None) as batch_op:
        batch_op.drop_column('try_number')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os
import tempfile
import unittest
from unittest.mock import patch

from ai_flow.util.file_util.zip_file_util import make_file_zipfile, make_dir_zipfile, extract_zip_file


class TestZipFileUtil(unittest.TestCase):

    def test_unzip_file(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            with open(os.path.join(tmp_dir, '1'), 'w') as f:
                f.write('abc')
            os.makedirs(os.path.join(tmp_dir, 'dir'))
            make_file_zipfile(os.path.join(tmp_dir, '1'), os.path.join(tmp_dir, '1.zip'))
            make_dir_zipfile(os.path.join(tmp_dir, 'dir'), os.path.join(tmp_dir, 'dir.zip'))

            self.assertTrue(os.path.exists(os.path.join(tmp_dir, '1.zip')))
            self.assertTrue(os.path.exists(os.path.join(tmp_dir, 'dir.zip')))

            with patch("zipfile.ZipFile.extractall") as zip_mock:
                extract_zip_file(os.path.join(tmp_dir, '1.zip'), os.path.join(tmp_dir, '1'))
                zip_mock.assert_not_called()
                extract_zip_file(os.path.join(tmp_dir, '1.zip'), os.path.join(tmp_dir, '2'))
                zip_mock.assert_called_once()

            with patch("zipfile.ZipFile.extractall") as zip_mock2:
                extract_zip_file(os.path.join(tmp_dir, 'dir.zip'), tmp_dir)
                zip_mock2.assert_not_called()
                extract_zip_file(os.path.join(tmp_dir, 'dir.zip'), os.path.join(tmp_dir, 'dir2'))
                zip_mock2.assert_called_once()





# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os
import shutil
import time
from typing import List

from pyflink.table import Table, DataTypes, ScalarFunction
from pyflink.table.udf import udf
from pyflink.version import __version__

from ai_flow_plugins.job_plugins import flink
from ai_flow_plugins.job_plugins.flink import INCLUDE_DATASET_VERSIONS
from ai_flow_plugins.job_plugins.flink.flink_processor import UDFWrapper

if __version__ in INCLUDE_DATASET_VERSIONS:
    from pyflink.table.descriptors import Schema, OldCsv, FileSystem


class PassUDF(ScalarFunction):

    def eval(self, s):
        return s


class Transformer(flink.FlinkPythonProcessor):
    def process(self, execution_context: flink.ExecutionContext, input_list: List[Table] = None) -> List[Table]:
        execution_context.table_env.register_function("pass_func", udf(PassUDF(),
                                                                       input_types=[DataTypes.STRING()],
                                                                       result_type=DataTypes.STRING()))
        return [input_list[0].group_by('word').select('pass_func(word), count(1)')]


class SleepUDF(ScalarFunction):

    def eval(self, s):
        time.sleep(100)
        return s


class Transformer2(flink.FlinkPythonProcessor):
    def process(self, execution_context: flink.ExecutionContext, input_list: List[Table] = None) -> List[Table]:
        execution_context.table_env.register_function("sleep_func", udf(SleepUDF(),
                                                                        input_types=[DataTypes.STRING()],
                                                                        result_type=DataTypes.STRING()))
        return [input_list[0].group_by('word').select('sleep_func(word), count(1)')]


class Source(flink.FlinkPythonProcessor):
    def process(self, execution_context: flink.ExecutionContext, input_list: List[Table] = None) -> List[Table]:
        input_file = os.path.join(os.getcwd(), 'resources', 'word_count.txt')
        t_env = execution_context.table_env
        if __version__ in INCLUDE_DATASET_VERSIONS:
            t_env.connect(FileSystem().path(input_file)) \
                .with_format(OldCsv()
                             .field('word', DataTypes.STRING())) \
                .with_schema(Schema()
                             .field('word', DataTypes.STRING())) \
                .create_temporary_table('mySource')
        else:
            t_env.execute_sql('''
                                            create table mySource (
                                                word string
                                            ) with (
                                                'connector' = 'filesystem',
                                                'path' = '{}',
                                                'format' = 'csv'
                                            )
                                        '''.format(input_file))
        return [t_env.from_path('mySource')]


class Sink(flink.FlinkPythonProcessor):
    def process(self, execution_context: flink.ExecutionContext, input_list: List[Table] = None) -> List[Table]:
        output_file = os.path.join(os.getcwd(), 'output')
        if os.path.exists(output_file):
            os.remove(output_file)

        t_env = execution_context.table_env
        statement_set = execution_context.statement_set
        if __version__ in INCLUDE_DATASET_VERSIONS:
            t_env.connect(FileSystem().path(output_file)) \
                .with_format(OldCsv()
                             .field_delimiter('\t')
                             .field('word', DataTypes.STRING())
                             .field('count', DataTypes.BIGINT())) \
                .with_schema(Schema()
                             .field('word', DataTypes.STRING())
                             .field('count', DataTypes.BIGINT())) \
                .create_temporary_table('mySink')
        else:
            t_env.execute_sql('''
                                                    create table mySink (
                                                        word string,
                                                        `count` bigint
                                                    ) with (
                                                        'connector' = 'filesystem',
                                                        'path' = '{}',
                                                        'format' = 'csv'
                                                    )
                                                '''.format(output_file))
        statement_set.add_insert('mySink', input_list[0])
        return []


class SinkWithExecuteSql(flink.FlinkPythonProcessor):
    def process(self,
                execution_context: flink.ExecutionContext,
                input_list: List[Table] = None) -> List[Table]:
        output_file = os.path.join(os.getcwd(), 'output')
        if os.path.exists(output_file):
            os.remove(output_file)

        t_env = execution_context.table_env
        if __version__ in INCLUDE_DATASET_VERSIONS:
            t_env.connect(FileSystem().path(output_file)) \
                .with_format(OldCsv()
                             .field_delimiter('\t')
                             .field('word', DataTypes.STRING())
                             .field('count', DataTypes.BIGINT())) \
                .with_schema(Schema()
                             .field('word', DataTypes.STRING())
                             .field('count', DataTypes.BIGINT())) \
                .create_temporary_table('mySink')
        else:
            t_env.execute_sql('''
                                        create table mySink (
                                            word string,
                                            `count` bigint
                                        ) with (
                                            'connector' = 'filesystem',
                                            'path' = '{}',
                                            'format' = 'csv'
                                        )
                                            '''.format(output_file))
        t_env.execute_sql(f"""
        INSERT INTO mySink
        SELECT *
        FROM {input_list[0]}
        """)
        return []


class SinkWithAddInsertSql(flink.FlinkPythonProcessor):
    def process(self,
                execution_context: flink.ExecutionContext,
                input_list: List[Table] = None) -> List[Table]:
        output_file = os.path.join(os.getcwd(), 'output')
        if os.path.exists(output_file):
            os.remove(output_file)

        t_env = execution_context.table_env
        s_set = execution_context.statement_set
        if __version__ in INCLUDE_DATASET_VERSIONS:
            t_env.connect(FileSystem().path(output_file)) \
                .with_format(OldCsv()
                             .field_delimiter('\t')
                             .field('word', DataTypes.STRING())
                             .field('count', DataTypes.BIGINT())) \
                .with_schema(Schema()
                             .field('word', DataTypes.STRING())
                             .field('count', DataTypes.BIGINT())) \
                .create_temporary_table('mySink')
        else:
            t_env.execute_sql('''
                                                    create table mySink (
                                                        word string,
                                                        `count` bigint
                                                    ) with (
                                                        'connector' = 'filesystem',
                                                        'path' = '{}',
                                                        'format' = 'csv'
                                                    )
                                                        '''.format(output_file))
        s_set.add_insert_sql(f"""
        INSERT INTO mySink
        SELECT *
        FROM {input_list[0]}
        """)
        return []


class SourceSql(flink.FlinkSqlProcessor):
    def __init__(self):
        super().__init__()
        self.input_file = None

    def open(self, execution_context: flink.ExecutionContext):
        self.input_file = os.path.join(os.getcwd(), 'resources', 'word_count.txt')

    def sql_statements(self, execution_context: flink.ExecutionContext) -> List[str]:
        table_source_stmt = '''
                           CREATE TABLE mySource (
                               word STRING 
                           ) WITH (
                               'connector' = 'filesystem',
                               'path' = '{uri}',
                               'format' = 'csv',
                               'csv.ignore-parse-errors' = 'true'
                           )
                            '''.format(uri=self.input_file)
        return [table_source_stmt]


class SinkSql(flink.FlinkSqlProcessor):
    def __init__(self):
        super().__init__()
        self.output_file = None

    def open(self, execution_context: flink.ExecutionContext):
        self.output_file = os.path.join(os.getcwd(), 'output')

        if os.path.exists(self.output_file):
            os.remove(self.output_file)
        if os.path.isdir(self.output_file):
            shutil.rmtree(self.output_file)

    def udf_list(self, execution_context: flink.ExecutionContext) -> List[UDFWrapper]:
        return [UDFWrapper(name='pass_func', func=udf(PassUDF(),
                                                      input_types=[DataTypes.STRING()],
                                                      result_type=DataTypes.STRING()))]

    def sql_statements(self, execution_context: flink.ExecutionContext) -> List[str]:
        table_sink_stmt = '''
                           CREATE TABLE mySink (
                               newword STRING
                           ) WITH (
                               'connector' = 'filesystem',
                               'path' = '{uri}',
                               'format' = 'csv',
                               'csv.ignore-parse-errors' = 'true'
                           )
        '''.format(uri=self.output_file)
        insert_stmt = 'INSERT INTO mySink SELECT pass_func(word) FROM mySource'
        return [table_sink_stmt, insert_stmt]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import signal
import time
import unittest
import os
import shutil
from subprocess import Popen

from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.utils.session import create_session
from airflow.utils.state import State
from notification_service.client import NotificationClient
from notification_service.base_notification import BaseEvent
from ai_flow import AIFlowServerRunner, init_ai_flow_context
from ai_flow.util.process_utils import get_all_children_pids
from ai_flow.workflow.status import Status
from ai_flow_plugins.job_plugins import bash
from ai_flow_plugins.tests.airflow_scheduler_utils import run_ai_flow_workflow, get_dag_id, \
    get_workflow_execution_info, set_workflow_execution_info, start_airflow_web_server
from ai_flow_plugins.tests import airflow_db_utils
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server
import ai_flow as af

project_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


def run_batch_workflow(client: NotificationClient):
    with af.job_config('task_1'):
        af.user_define_operation(processor=bash.BashProcessor(bash_command='echo "Xiao ming hello world!"'))
    workflow_info = af.workflow_operation.submit_workflow(
        workflow_name=af.current_workflow_config().workflow_name)
    wei = af.workflow_operation.start_new_workflow_execution(
        workflow_name=af.current_workflow_config().workflow_name)
    set_workflow_execution_info(wei)
    while True:
        with create_session() as session:
            dag_run = session.query(DagRun) \
                .filter(DagRun.dag_id == 'test_project.{}'
                        .format(af.current_workflow_config().workflow_name)).first()
            if dag_run is not None and dag_run.state == State.SUCCESS:
                break
            else:
                time.sleep(1)


def run_event_workflow(client: NotificationClient):
    with af.job_config('task_1'):
        af.user_define_operation(processor=bash.BashProcessor(bash_command='echo "Xiao ming hello world!"'))
    af.action_on_event(job_name='task_1',
                       event_key='test_key',
                       event_type='test_type',
                       event_value='test_value',
                       sender='*',
                       namespace='*',
                       action=af.JobAction.RESTART)
    workflow_info = af.workflow_operation.submit_workflow(
        workflow_name=af.current_workflow_config().workflow_name)
    wei = af.workflow_operation.start_new_workflow_execution(
        workflow_name=af.current_workflow_config().workflow_name)
    set_workflow_execution_info(wei)
    result = af.workflow_operation.start_job_execution(
        job_name='task_1',
        execution_id=get_workflow_execution_info().workflow_execution_id)

    while True:
        with create_session() as session:
            task_instance = session.query(TaskInstance) \
                .filter(TaskInstance.dag_id == 'test_project.{}'
                        .format(af.current_workflow_config().workflow_name),
                        TaskInstance.task_id == 'task_1'
                        ).first()
            if task_instance is not None and task_instance.state == State.SUCCESS:
                break
            else:
                time.sleep(1)

    af.workflow_operation.stop_workflow_execution(wei.workflow_execution_id)

    while True:
        with create_session() as session:

            dag_run = session.query(DagRun) \
                .filter(DagRun.dag_id == 'test_project.{}'
                        .format(af.current_workflow_config().workflow_name),
                        ).first()
            if dag_run is not None and dag_run.state == State.KILLED:
                break
            else:
                time.sleep(1)


class TestBash(unittest.TestCase):
    master = None
    web_process = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.ns_server = start_notification_server()
        airflow_path = '/tmp/airflow_deploy'
        if os.path.exists(airflow_path):
            shutil.rmtree(airflow_path)
        os.makedirs(airflow_path)

        config_file = project_path + '/master.yaml'
        cls.master = AIFlowServerRunner(config_file=config_file)
        cls.master.start()

        cls.web_process: Popen = start_airflow_web_server()

    @classmethod
    def tearDownClass(cls) -> None:
        sub_pids = get_all_children_pids(cls.web_process.pid)
        print(sub_pids)
        for pid in sub_pids:
            os.kill(pid, signal.SIGKILL)
        cls.web_process.kill()
        time.sleep(5)
        cls.master.stop()
        stop_notification_server(cls.ns_server)

    def setUp(self):
        airflow_db_utils.clear_all()
        self.master._clear_db()
        af.current_graph().clear_graph()
        init_ai_flow_context()

    def tearDown(self):
        self.master._clear_db()
        generated = '{}/generated'.format(project_path)
        if os.path.exists(generated):
            shutil.rmtree(generated)
        temp = '{}/temp'.format(project_path)
        if os.path.exists(temp):
            shutil.rmtree(temp)

    def test_workflow_apis(self):
        run_ai_flow_workflow(dag_id=get_dag_id(af.current_project_config().get_project_name(),
                                               af.current_workflow_config().workflow_name),
                             test_function=run_batch_workflow)

        result = af.workflow_operation.list_workflows(offset=0, page_size=5)
        self.assertEqual(1, len(result))

        result = af.workflow_operation.get_workflow(workflow_name='test_bash')
        self.assertTrue(result is not None)

        result = af.workflow_operation.list_workflow_executions(workflow_name='test_bash')
        self.assertEqual(1, len(result))

        result = af.workflow_operation.get_workflow_execution(
            execution_id=get_workflow_execution_info().workflow_execution_id)
        self.assertTrue(result is not None)
        self.assertEqual(Status.FINISHED, result.status)

        job_execution_infos = af.workflow_operation.get_job_executions(job_name='task_1',
                                                                       execution_id=get_workflow_execution_info().
                                                                       workflow_execution_id)
        self.assertEqual(Status.FINISHED, job_execution_infos[0].status)

    def test_event_workflow_apis(self):
        run_ai_flow_workflow(dag_id=get_dag_id(af.current_project_config().get_project_name(),
                                               af.current_workflow_config().workflow_name),
                             test_function=run_event_workflow)

        result = af.workflow_operation.get_workflow_execution(
            execution_id=get_workflow_execution_info().workflow_execution_id)
        self.assertTrue(result is not None)
        self.assertEqual(Status.KILLED, result.status)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import unittest

import ai_flow as af
from ai_flow.api import ai_flow_context
from ai_flow.api.ai_flow_context import init_ai_flow_context, init_notebook_context
from ai_flow.context.project_context import current_project_config, current_project_context
from ai_flow.context.workflow_config_loader import current_workflow_config
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.project.project_config import ProjectConfig
from ai_flow.store.db.base_model import base
from ai_flow.test.store.test_sqlalchemy_store import _get_store
from ai_flow.test.util.notification_service_utils import start_notification_server, stop_notification_server, _NS_URI
from ai_flow.test.util.server_util import wait_for_server_started
from ai_flow.util import sqlalchemy_db
from ai_flow.workflow.job_config import JobConfig
from ai_flow.workflow.workflow_config import WorkflowConfig

_SQLITE_DB_FILE = 'aiflow.db'
_SQLITE_DB_URI = '%s%s' % ('sqlite:///', _SQLITE_DB_FILE)
_PORT = '50051'


class TestAIFlowContext(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        print("TestAIFlowContext setUpClass")
        cls.ns_server = start_notification_server()
        if os.path.exists(_SQLITE_DB_FILE):
            os.remove(_SQLITE_DB_FILE)
        cls.server = AIFlowServer(store_uri=_SQLITE_DB_URI, port=_PORT, start_scheduler_service=False)
        cls.server.run()
        wait_for_server_started('localhost:{}'.format(_PORT))

    @classmethod
    def tearDownClass(cls) -> None:
        cls.server.stop()
        os.remove(_SQLITE_DB_FILE)
        stop_notification_server(cls.ns_server)

    def setUp(self) -> None:
        _get_store(_SQLITE_DB_URI)

    def tearDown(self) -> None:
        sqlalchemy_db.clear_db(_SQLITE_DB_URI, base.metadata)
        ai_flow_context.__init_context_flag__ = False
        ai_flow_context.__init_notebook_context_flag__ = False
        ai_flow_context.__init_client_flag__ = False

    def test_init_notebook_context(self):
        project_config: ProjectConfig = ProjectConfig()
        project_config.set_project_name('test_project')
        project_config.set_server_uri('localhost:50051')
        project_config.set_notification_server_uri('localhost:50052')
        project_config['blob'] = {
            'blob_manager_class': 'ai_flow.test.api.mock_plugins.MockBlobManger'}
        project_config['a'] = 'a'

        workflow_config: WorkflowConfig = WorkflowConfig('test_notebook_context')
        workflow_config.add_job_config('task_1', JobConfig(job_name='task_1', job_type='mock'))
        workflow_config.add_job_config('task_2', JobConfig(job_name='task_2', job_type='mock'))
        init_notebook_context(project_config, workflow_config)
        self.assertEqual('test_project', current_project_config().get_project_name())
        self.assertEqual('a', current_project_config().get('a'))
        self.assertEqual('test_project', current_project_context().project_name)
        self.assertEqual('test_notebook_context', current_workflow_config().workflow_name)
        self.assertEqual(2, len(current_workflow_config().job_configs))

        init_notebook_context(project_config, workflow_config)
        self.assertEqual('test_project', current_project_config().get_project_name())

        with self.assertRaises(Exception):
            init_ai_flow_context()

    def test_init_ai_client(self):
        af.init_ai_flow_client(server_uri='localhost:50051', project_name='test', notification_server_uri=_NS_URI)
        project_meta = af.get_project_by_name('test')
        self.assertEqual('test', project_meta.name)

    def test_init_ai_client_no_project_name(self):
        af.init_ai_flow_client(server_uri='localhost:50051', project_name=None, notification_server_uri=_NS_URI)
        project_meta = af.get_project_by_name('Unknown')
        self.assertEqual('Unknown', project_meta.name)

    def test_init_ai_client_register_dataset(self):
        af.init_ai_flow_client(server_uri='localhost:50051', project_name=None, notification_server_uri=_NS_URI)
        af.register_dataset(name='test_dataset', uri='/test')
        dataset_meta = af.get_dataset_by_name('test_dataset')
        self.assertEqual('/test', dataset_meta.uri)


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import asyncio
import functools
import inspect
import logging
import os
import sys
import tempfile
import threading
import time
from concurrent import futures
from typing import Dict

import grpc
from grpc import _common, _server
from grpc._cython.cygrpc import StatusCode
from grpc._server import _serialize_response, _status, _abort, _Context, _unary_request, \
    _select_thread_pool_for_behavior, _unary_response_in_pool

from ai_flow.endpoint.server.high_availability import SimpleAIFlowServerHaManager, HighAvailableService
from ai_flow.endpoint.server.server_config import DBType
from ai_flow.metadata_store.service.service import MetadataService
from ai_flow.metric.service.metric_service import MetricService
from ai_flow.model_center.service.service import ModelCenterService
from ai_flow.protobuf.high_availability_pb2_grpc import add_HighAvailabilityManagerServicer_to_server
from ai_flow.scheduler_service.service.service import SchedulerService, SchedulerServiceConfig
from ai_flow.settings import AIFLOW_HOME
from ai_flow.store.db.base_model import base
from ai_flow.store.db.db_util import extract_db_engine_from_uri, create_db_store
from ai_flow.store.mongo_store import MongoStoreConnManager
from ai_flow.store.sqlalchemy_store import SqlAlchemyStore
from ai_flow.util import sqlalchemy_db
from notification_service.proto import notification_service_pb2_grpc
from notification_service.service import NotificationService

sys.path.append(os.path.abspath(os.path.join(os.getcwd(), "../../..")))

from ai_flow.protobuf import model_center_service_pb2_grpc, \
    metadata_service_pb2_grpc, metric_service_pb2_grpc, scheduling_service_pb2_grpc

_PORT = '50051'
_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class AIFlowServer(object):
    """
    Block/Async server of an AIFlow Rest Endpoint that provides Metadata/Model/Notification function service.
    """

    def __init__(self,
                 store_uri=None,
                 port=_PORT,
                 notification_server_uri=None,
                 start_meta_service: bool = True,
                 start_model_center_service: bool = True,
                 start_metric_service: bool = True,
                 start_scheduler_service: bool = True,
                 scheduler_service_config: Dict = None,
                 enabled_ha: bool = False,
                 ha_manager=None,
                 ha_server_uri=None,
                 ha_storage=None,
                 ttl_ms: int = 10000,
                 base_log_folder: str = AIFLOW_HOME):
        self.store_uri = store_uri
        self.db_type = DBType.value_of(extract_db_engine_from_uri(store_uri))
        self.executor = Executor(futures.ThreadPoolExecutor(max_workers=10))
        self.server = grpc.server(self.executor)
        self.enabled_ha = enabled_ha
        self.start_scheduler_service = start_scheduler_service
        server_uri = 'localhost:{}'.format(port)

        if start_model_center_service:
            logging.info("start model center service.")

            model_center_service_pb2_grpc.add_ModelCenterServiceServicer_to_server(
                ModelCenterService(store_uri=store_uri), self.server)
        if start_meta_service:
            logging.info("start meta service.")
            metadata_service_pb2_grpc.add_MetadataServiceServicer_to_server(
                MetadataService(db_uri=store_uri, server_uri=server_uri), self.server)
        if start_metric_service:
            logging.info("start metric service.")
            metric_service_pb2_grpc.add_MetricServiceServicer_to_server(MetricService(db_uri=store_uri), self.server)

        if start_scheduler_service:
            self._add_scheduler_service(scheduler_service_config, store_uri, notification_server_uri, base_log_folder)

        if enabled_ha:
            self._add_ha_service(ha_manager, ha_server_uri, ha_storage, store_uri, ttl_ms)

        self.server.add_insecure_port('[::]:' + str(port))

        self._stop = threading.Event()

    def _add_scheduler_service(self, scheduler_service_config, db_uri, notification_server_uri, base_log_folder):
        logging.info("start scheduler service.")
        real_config = SchedulerServiceConfig(scheduler_service_config)
        self.scheduler_service = SchedulerService(real_config, db_uri, notification_server_uri, base_log_folder)
        scheduling_service_pb2_grpc.add_SchedulingServiceServicer_to_server(self.scheduler_service,
                                                                            self.server)

    def _add_ha_service(self, ha_manager, ha_server_uri, ha_storage, store_uri, ttl_ms):
        if ha_manager is None:
            ha_manager = SimpleAIFlowServerHaManager()
        if ha_server_uri is None:
            raise ValueError("ha_server_uri is required with ha enabled!")
        if ha_storage is None:
            ha_storage = create_db_store(store_uri)
        self.ha_service = HighAvailableService(ha_manager, ha_server_uri, ha_storage, ttl_ms)
        add_HighAvailabilityManagerServicer_to_server(self.ha_service, self.server)

    def run(self, is_block=False):
        if self.enabled_ha:
            self.ha_service.start()
        self.server.start()
        if self.start_scheduler_service:
            self.scheduler_service.start()
        logging.info('AIFlow server started.')
        if is_block:
            try:
                while not self._stop.is_set():
                    self._stop.wait(_ONE_DAY_IN_SECONDS)
            except KeyboardInterrupt:
                logging.info("received KeyboardInterrupt")
                self.stop()
        else:
            pass

    def stop(self, clear_sql_lite_db_file=False):
        logging.info("stopping AIFlow server")
        if self.start_scheduler_service:
            self.scheduler_service.stop()
        self.server.stop(0)
        if self.enabled_ha:
            self.ha_service.stop()
        self.executor.shutdown()

        if self.db_type == DBType.SQLITE and clear_sql_lite_db_file:
            sqlalchemy_db.clear_db(self.store_uri, base.metadata)
            os.remove(self.store_uri[10:])
        elif self.db_type == DBType.MONGODB:
            MongoStoreConnManager().disconnect_all()

        self._stop.set()
        logging.info('AIFlow server stopped.')

    def _clear_db(self):
        if self.db_type == DBType.SQLITE:
            sqlalchemy_db.reset_db(self.store_uri, base.metadata)
        elif self.db_type == DBType.MONGODB:
            MongoStoreConnManager().drop_all()


def _loop(loop: asyncio.AbstractEventLoop):
    asyncio.set_event_loop(loop)
    if not loop.is_running() or loop.is_closed():
        loop.run_forever()
    pending = asyncio.all_tasks(loop=loop)
    if pending:
        loop.run_until_complete(asyncio.gather(*pending))


class Executor(futures.Executor):
    def __init__(self, thread_pool, loop=None):
        super().__init__()
        self._shutdown = False
        self._thread_pool = thread_pool
        self._loop = loop or asyncio.get_event_loop()
        if not self._loop.is_running() or self._loop.is_closed():
            self._thread = threading.Thread(target=_loop, args=(self._loop,), daemon=True)
            self._thread.start()

    def submit(self, fn, *args, **kwargs):
        if self._shutdown:
            raise RuntimeError('Cannot schedule new futures after shutdown.')
        if not self._loop.is_running():
            raise RuntimeError('Loop must be started before any function could be submitted.')
        if inspect.iscoroutinefunction(fn):
            coroutine = fn(*args, **kwargs)
            return asyncio.run_coroutine_threadsafe(coroutine, self._loop)
        else:
            func = functools.partial(fn, *args, **kwargs)
            return self._loop.run_in_executor(self._thread_pool, func)

    def shutdown(self, wait=True):
        self._shutdown = True
        if wait:
            self._thread_pool.shutdown()


async def _call_behavior_async(rpc_event, state, behavior, argument, request_deserializer):
    context = _Context(rpc_event, state, request_deserializer)
    try:
        return await behavior(argument, context), True
    except Exception as e:
        with state.condition_type:
            if e not in state.rpc_errors:
                logging.exception(e)
                _abort(state, rpc_event.operation_call, StatusCode.unknown, _common.encode(e))
        return None, False


async def _unary_response_in_pool_async(rpc_event, state, behavior, argument_thunk, request_deserializer,
                                        response_serializer):
    argument = argument_thunk()
    if argument is not None:
        response, proceed = await _call_behavior_async(rpc_event, state, behavior, argument, request_deserializer)
        if proceed:
            serialized_response = _serialize_response(rpc_event, state, response, response_serializer)
            if serialized_response is not None:
                _status(rpc_event, state, serialized_response)


def _handle_unary_unary(rpc_event, state, method_handler, default_thread_pool):
    unary_request = _unary_request(rpc_event, state, method_handler.request_deserializer)
    thread_pool = _select_thread_pool_for_behavior(method_handler.unary_unary, default_thread_pool)
    if asyncio.iscoroutinefunction(method_handler.unary_unary):
        return thread_pool.submit(_unary_response_in_pool_async, rpc_event, state, method_handler.unary_unary,
                                  unary_request, method_handler.request_deserializer,
                                  method_handler.response_serializer)
    else:
        return thread_pool.submit(_unary_response_in_pool, rpc_event, state, method_handler.unary_unary, unary_request,
                                  method_handler.request_deserializer, method_handler.response_serializer)


_server._handle_unary_unary = _handle_unary_unary

if __name__ == '__main__':
    fd, temp_db_file = tempfile.mkstemp()
    os.close(fd)
    store_uri = '%s%s' % ('sqlite:///', temp_db_file)
    server = AIFlowServer(store_uri=store_uri)
    server.run(is_block=True)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add_workflow_snapshot

Revision ID: 0e8fefcce376
Revises: de045f309236
Create Date: 2021-11-24 10:08:58.313703

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0e8fefcce376'
down_revision = 'de045f309236'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('workflow_snapshot',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('workflow_id', sa.BigInteger(), nullable=False),
                    sa.Column('uri', sa.String(length=1000), nullable=True),
                    sa.Column('signature', sa.String(length=1000), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.ForeignKeyConstraint(['workflow_id'], ['workflow.uuid'], onupdate='cascade'),
                    )


def downgrade():
    op.drop_table('workflow_snapshot')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import threading
from abc import ABC, abstractmethod
from typing import Text, Set

import typing

from ai_flow.ai_graph.ai_node import ReadDatasetNode, WriteDatasetNode

from ai_flow.plugin_interface.scheduler_interface import JobExecutionInfo

from ai_flow.workflow.job_config import JobConfig

from ai_flow.ai_graph.ai_graph import AISubGraph
from ai_flow.runtime.job_runtime_env import JobRuntimeEnv
from ai_flow.translator.translator import JobGenerator

from ai_flow.plugin_interface.job_plugin_interface import JobHandle, JobController, JobPluginFactory

from ai_flow.workflow.job import Job
from ai_flow.workflow.status import Status
from ai_flow_plugins.job_plugins.read_only.read_only_processor import ReadOnlyProcessor


class ReadOnlyJob(Job, ABC):
    def __init__(self, job_config: JobConfig):
        super().__init__(job_config)


class ReadOnlyJobHandle(JobHandle):
    def __init__(self, job: Job, job_execution_info: JobExecutionInfo):
        super().__init__(job, job_execution_info)


class ReadOnlyJobGenerator(JobGenerator):
    def __init__(self, required_properties=None):

        if required_properties is None:
            required_properties = set()
        self._required_properties = required_properties

    def generate(self, sub_graph: AISubGraph, resource_dir: Text = None) -> Job:
        for node_name, node in sub_graph.nodes.items():
            if isinstance(node, ReadDatasetNode) or isinstance(node, WriteDatasetNode):
                continue
            processor = node.get_processor()
            if not isinstance(processor, ReadOnlyProcessor):
                raise TypeError("Getting unknown processor type {} for node {}".format(type(processor), node_name))
        job_config = sub_graph.config
        self._validate_job_config(job_config)
        return ReadOnlyJob(job_config)

    def _validate_job_config(self, job_config: JobConfig):
        missing_properties = []
        for required_properties in self._required_properties:
            if required_properties not in job_config.properties:
                missing_properties.append(required_properties)

        if len(missing_properties) != 0:
            raise RuntimeError("Missing required properties: {}".format(missing_properties))


class ReadOnlyJobController(JobController):
    """
    ReadOnlyJobController implement JobController to manage the life cycle of a ReadOnlyJob. The ReadOnlyJobController
    is used to register a job to the workflow without actually submitting the job. It can report some runtime
    status of the job, if its get_job_label method is overwritten.

    This is useful when you want to register a job to the workflow so that you can capture the job's runtime status.
    And you don't want the workflow to actually run and stop a job.
    """

    def __init__(self):
        super().__init__()
        self._job_stop_events: typing.Dict[Job, threading.Event] = {}

    def submit_job(self, job: Job, job_runtime_env: JobRuntimeEnv) -> JobHandle:
        # submit job do not actually submit a job, it only return the JobHandle
        self._job_stop_events[job] = threading.Event()
        return ReadOnlyJobHandle(job, job_runtime_env.job_execution_info)

    def stop_job(self, job_handle: JobHandle, job_runtime_env: JobRuntimeEnv):
        self._check_job_exists(job_handle)
        self._job_stop_events[job_handle.job].set()

    def cleanup_job(self, job_handle: JobHandle, job_runtime_env: JobRuntimeEnv):
        # do nothing
        pass

    def get_result(self, job_handle: JobHandle, blocking: bool = True) -> object:
        self._check_job_exists(job_handle)

        if not blocking:
            return None

        self._job_stop_events[job_handle.job].wait()
        return None

    def get_job_status(self, job_handle: JobHandle) -> Status:
        self._check_job_exists(job_handle)
        # The read only job itself should always be running
        return Status.RUNNING

    def obtain_job_label(self, job_handle: JobHandle) -> Text:
        self._check_job_exists(job_handle)
        if not isinstance(job_handle.job, ReadOnlyJob):
            raise TypeError("Unknown job type: {}".format(type(job_handle.job)))
        return self.get_job_label(job_handle.job)

    def get_job_label(self, job: ReadOnlyJob) -> Text:
        return ""

    def _check_job_exists(self, job_handle):
        if job_handle.job not in self._job_stop_events:
            raise RuntimeError("Cannot find job {}.".format(job_handle.job))


class ReadOnlyJobPluginFactory(JobPluginFactory):
    def job_type(self) -> Text:
        return "read_only_job"

    def get_job_generator(self) -> JobGenerator:
        return ReadOnlyJobGenerator()

    def get_job_controller(self) -> JobController:
        return ReadOnlyJobController()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import time
import unittest
from unittest import mock

from ai_flow.runtime.job_runtime_env import JobRuntimeEnv
from ai_flow.workflow.job import Job
from ai_flow.workflow.job_config import JobConfig

from ai_flow.workflow.workflow import Workflow, WorkflowPropertyKeys

from ai_flow.workflow.status import Status

from ai_flow.plugin_interface.job_plugin_interface import JobController, JobHandle
from ai_flow_plugins.job_plugins.read_only import ReadOnlyJob
from airflow import AirflowException

from airflow.models.taskexecution import TaskExecution

from ai_flow_plugins.scheduler_plugins.airflow.ai_flow_operator import ExecutionLabelReportThread, AIFlowOperator


class TestingJobController(JobController):

    def submit_job(self, job: Job, job_runtime_env: JobRuntimeEnv) -> JobHandle:
        pass

    def stop_job(self, job_handle: JobHandle, job_runtime_env: JobRuntimeEnv):
        pass

    def cleanup_job(self, job_handle: JobHandle, job_runtime_env: JobRuntimeEnv):
        pass

    def get_result(self, job_handle: JobHandle, blocking: bool = True) -> object:
        pass

    def get_job_status(self, job_handle: JobHandle) -> Status:
        pass


class TestAIFlowOperator(unittest.TestCase):

    def setUp(self) -> None:
        workflow = mock.MagicMock()
        workflow.properties = dict()
        workflow.properties[WorkflowPropertyKeys.JOB_PLUGINS] = \
            {'test_job': (TestingJobController.__module__, TestingJobController.__name__)}
        job_config = JobConfig()
        job_config.job_type = 'test_job'
        job = ReadOnlyJob(job_config)
        self.op = AIFlowOperator(workflow=workflow, job=job, task_id='1')

    def test_execute(self):
        with mock.patch.object(self.op, 'job_controller', wraps=self.op.job_controller) as job_controller:
            job_controller.get_result.return_value = 'result'
            mock_handle = mock.MagicMock()
            job_controller.submit_job.return_value = mock_handle
            self.assertEqual('result', self.op.execute({}))
            job_controller.submit_job.assert_called_once()
            job_controller.stop_job.assert_not_called()
            job_controller.get_result.assert_called_once_with(job_handle=mock_handle, blocking=True)
            job_controller.cleanup_job.assert_called_once_with(mock_handle, mock.ANY)

    def test_execute_raise_AirflowException(self):
        with mock.patch.object(self.op, 'job_controller', wraps=self.op.job_controller) as job_controller:
            job_controller.get_result.side_effect = AirflowException("Boom! Shakalaka!")
            mock_handle = mock.MagicMock()
            job_controller.submit_job.return_value = mock_handle

            with self.assertRaises(AirflowException):
                self.assertEqual(None, self.op.execute({}))
                job_controller.submit_job.assert_called_once()
                job_controller.get_result.assert_called_once_with(job_handle=mock_handle, blocking=True)
                job_controller.stop_job.assert_called_once_with(mock_handle, mock.ANY)
                job_controller.cleanup_job.assert_called_once_with(mock_handle, mock.ANY)

    def test_execute_raise_unexpected_exception(self):
        with mock.patch.object(self.op, 'job_controller', wraps=self.op.job_controller) as job_controller:
            job_controller.get_result.side_effect = RuntimeError("Boom! Shakalaka!")
            mock_handle = mock.MagicMock()
            job_controller.submit_job.return_value = mock_handle

            with self.assertRaises(RuntimeError):
                self.assertEqual(None, self.op.execute({}))
                job_controller.submit_job.assert_called_once()
                job_controller.get_result.assert_called_once_with(job_handle=mock_handle, blocking=True)
                job_controller.stop_job.assert_called_once_with(mock_handle, mock.ANY)
                job_controller.cleanup_job.assert_called_once_with(mock_handle, mock.ANY)


class TestExecutionLabelReportThread(unittest.TestCase):

    def test_run(self):
        task_execution: TaskExecution = mock.Mock()
        job_controller: JobController = mock.Mock()
        job_controller.obtain_job_label.return_value = 'test_label'
        job_controller.get_job_status.return_value = Status.RUNNING
        job_handle = mock.Mock()

        t = ExecutionLabelReportThread(task_execution, 5.0, job_controller, job_handle)

        try:
            t.start()
            time.sleep(1)
            task_execution.update_task_execution_label.assert_called_with('test_label')
        finally:
            t.stop()
            t.join()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import threading
import os
from typing import Any, Text

from airflow.exceptions import AirflowException
from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults
from ai_flow.util.file_util.zip_file_util import extract_zip_file
from ai_flow.workflow.status import Status
from ai_flow.common.module_load import import_string
from ai_flow.runtime.job_runtime_util import prepare_job_runtime_env
from ai_flow.util.time_utils import datetime_to_int64
from ai_flow.plugin_interface.blob_manager_interface import BlobConfig
from ai_flow.plugin_interface.job_plugin_interface import JobController, JobHandle, JobRuntimeEnv
from ai_flow.context.project_context import build_project_context
from ai_flow.plugin_interface.blob_manager_interface import BlobManagerFactory
from ai_flow.plugin_interface.scheduler_interface import JobExecutionInfo, WorkflowExecutionInfo, WorkflowInfo
from ai_flow.workflow.job import Job
from ai_flow.workflow.workflow import Workflow, WorkflowPropertyKeys
from ai_flow_plugins.scheduler_plugins.airflow.airflow_scheduler import AirFlowSchedulerBase


class StoppableThread(threading.Thread):
    """
    Thread class with a stop() method. The thread itself has to check regularly for the stopped() condition.
    """

    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._ended = threading.Event()

    def stop(self):
        self._ended.set()

    def stopped(self):
        return self._ended.is_set()


class ExecutionLabelReportThread(StoppableThread):
    """
    A thread that reports the execution label of a job periodically.
    """

    def __init__(self, task_execution, label_update_interval, job_controller, job_handle, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._task_execution = task_execution
        self._job_controller = job_controller
        self._job_handle = job_handle
        self._label_update_interval = label_update_interval

    def run(self) -> None:
        old_labels = None
        while not self.stopped() and self._job_controller.get_job_status(self._job_handle) == Status.RUNNING:
            label = self._job_controller.obtain_job_label(self._job_handle)
            if label != old_labels and len(label) > 0:
                self._task_execution.update_task_execution_label(label)
                old_labels = label
            self._ended.wait(self._label_update_interval)


class AIFlowOperator(BaseOperator):
    """
    AIFlowOperator implements airflow Operator.
    Its function is to submit and stop the job defined by the ai flow program.
    """

    @apply_defaults
    def __init__(
            self,
            job: Job,
            workflow: Workflow,
            resource_dir: Text = None,
            base_log_folder: Text = None,
            **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.job: Job = job
        self.workflow: Workflow = workflow
        plugins = self.workflow.properties.get(WorkflowPropertyKeys.JOB_PLUGINS)
        module, name = plugins.get(self.job.job_config.job_type)
        class_object = import_string('{}.{}'.format(module, name))
        self.job_controller: JobController = class_object()
        self.job_handle: JobHandle = None
        self.job_runtime_env: JobRuntimeEnv = None
        self.resource_dir = resource_dir
        self.base_log_folder = base_log_folder

    def context_to_job_info(self, project_name: Text, context: Any) -> JobExecutionInfo:
        """
        The function of this function is to turn the context of airflow into execution information of a job.
        """
        wi = WorkflowInfo(namespace=project_name, workflow_name=self.workflow.workflow_name)
        we = WorkflowExecutionInfo(workflow_execution_id=
                                   AirFlowSchedulerBase.generate_workflow_execution_id(
                                       dag_id=context.get('dag').dag_id,
                                       run_id=context.get('dag_run').run_id
                                   ),
                                   workflow_info=wi,
                                   start_date=str(datetime_to_int64(context.get('dag_run').start_date)),
                                   end_date=str(datetime_to_int64(context.get('dag_run').end_date)),
                                   status=context.get('dag_run').get_state(),
                                   context=context.get('dag_run').context)
        je = JobExecutionInfo(job_name=self.job.job_name,
                              job_execution_id=AirFlowSchedulerBase.generate_job_execution_id(
                                  dag_id=context.get('dag').dag_id,
                                  run_id=context.get('dag_run').run_id,
                                  task_id=context.get('ti').task_id,
                                  seq_num=str(context.get('te').seq_num),
                              ),
                              status=context.get('ti').state,
                              start_date=str(datetime_to_int64(context.get('ti').start_date)),
                              end_date=str(datetime_to_int64(context.get('ti').end_date)),
                              workflow_execution=we)
        return je

    def pre_execute(self, context: Any):
        config = {}
        config.update(self.workflow.properties['blob'])
        blob_config = BlobConfig(config)

        workflow_repo = os.path.join(self.resource_dir, self.workflow.workflow_name)
        if not os.path.exists(workflow_repo):
            os.makedirs(workflow_repo)
        blob_manager = BlobManagerFactory.create_blob_manager(blob_config.blob_manager_class(),
                                                              blob_config.blob_manager_config())
        project_zip_path = blob_manager.download(remote_file_path=self.workflow.project_uri,
                                                 local_dir=workflow_repo)
        extract_path = os.path.splitext(project_zip_path)[0]
        project_path: Text = extract_zip_file(zip_file_path=project_zip_path,
                                              extract_path=extract_path)

        self.log.info("project_path:" + project_path)
        project_context = build_project_context(project_path)

        job_execution_info: JobExecutionInfo = self.context_to_job_info(project_context.project_name, context)
        if context['conf'].has_option('scheduler', 'working_dir'):
            root_working_dir = context['conf'].get('scheduler', 'working_dir')
        else:
            root_working_dir = os.path.join(project_context.project_path, 'temp')
        self.log.info('working dir: {}'.format(root_working_dir))

        workflow_generated_dir = self.workflow.properties.get(WorkflowPropertyKeys.GENERATED_DIR)
        self.job_runtime_env = prepare_job_runtime_env(workflow_generated_dir=workflow_generated_dir,
                                                       workflow_name=self.workflow.workflow_name,
                                                       project_context=project_context,
                                                       job_execution_info=job_execution_info,
                                                       root_working_dir=root_working_dir,
                                                       base_log_folder=self.base_log_folder)

    def execute(self, context: Any):
        self.log.info("context:" + str(context))
        execution_label_report_thread = None
        result = None
        try:
            self.log.info("submitting job with job_runtime_env: {}".format(self.job_runtime_env))
            self.job_handle: JobHandle = self.job_controller.submit_job(self.job, self.job_runtime_env)

            if 'te' in context:
                self.log.info("Starting execution label report thread.")
                task_execution = context['te']
                execution_label_report_thread = \
                    ExecutionLabelReportThread(task_execution,
                                               self.job.job_config.job_label_report_interval,
                                               self.job_controller,
                                               self.job_handle)
                execution_label_report_thread.start()
            else:
                self.log.warning("te is not in context, execution label will not be reported.")

            result = self.job_controller.get_result(job_handle=self.job_handle, blocking=True)
        except AirflowException as e:
            self.log.warning("AirflowException during executing", exc_info=e)
            self.job_controller.stop_job(self.job_handle, self.job_runtime_env)
            raise e
        except Exception as e:
            self.job_controller.stop_job(self.job_handle, self.job_runtime_env)
            raise RuntimeError("Unexpected exception during executing") from e
        finally:
            if execution_label_report_thread:
                self.log.info("Stopping execution label report thread.")
                execution_label_report_thread.stop()
                execution_label_report_thread.join()
            self.job_controller.cleanup_job(self.job_handle, self.job_runtime_env)
        return result

    def on_kill(self):
        try:
            self.job_controller.stop_job(self.job_handle, self.job_runtime_env)
        finally:
            self.job_controller.cleanup_job(self.job_handle, self.job_runtime_env)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import pickle
import queue
import pytz
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.date import DateTrigger
from notification_service.base_notification import BaseEvent

from airflow.models.event_progress import EventProgress
from airflow.utils import timezone
from airflow.utils.session import provide_session
from airflow.models.message import Message, IdentifiedMessage, MessageState
from airflow.utils.log.logging_mixin import LoggingMixin


def put_message_to_queue(queue, message):
    queue.put(message)


class Mailbox(LoggingMixin):

    def __init__(self) -> None:
        super().__init__()
        self.queue = queue.Queue()
        self.scheduling_job_id = None
        self.sc = BackgroundScheduler()

    def start(self):
        self.sc.start()

    def stop(self):
        self.sc.shutdown()

    @provide_session
    def _save_message_to_db(self, message, queue_time, session=None) -> IdentifiedMessage:
        """ 1. save message to db
            2. update the event progress
        """
        try:
            if isinstance(message, BaseEvent) and message.version is not None and message.create_time is not None:
                progress = EventProgress(scheduling_job_id=self.scheduling_job_id,
                                         last_event_time=message.create_time,
                                         last_event_version=message.version)
                session.merge(progress)

            message_obj = Message(message)
            message_obj.state = MessageState.QUEUED
            message_obj.scheduling_job_id = self.scheduling_job_id
            message_obj.queue_time = queue_time
            session.add(message_obj)
            session.commit()
            return IdentifiedMessage(serialized_message=message_obj.data,
                                     msg_id=message_obj.id,
                                     queue_time=message_obj.queue_time)
        except Exception as e:
            session.rollback()
            raise e

    def send_message(self, message, queue_time=timezone.utcnow()):
        if not self.scheduling_job_id:
            self.log.warning("scheduling_job_id not set, missing messages cannot be recovered.")
        identified_message = self._save_message_to_db(message, queue_time)
        current_time = timezone.utcnow()
        if current_time >= queue_time:
            self.queue.put(identified_message)
        else:
            self.sc.add_job(func=put_message_to_queue,
                            args=(self.queue, identified_message),
                            trigger=DateTrigger(run_date=queue_time,
                                                timezone=pytz.timezone('UTC')))

    def get_message(self):
        identified_message: IdentifiedMessage = self.queue.get()
        try:
            return pickle.loads(identified_message.serialized_message)
        except Exception as e:
            self.log.error("Error occurred when load message from database, %s", e)
            return None

    def get_identified_message(self) -> IdentifiedMessage:
        return self.get_message_with_timeout(timeout=1)

    def length(self):
        return self.queue.qsize()

    def get_message_with_timeout(self, timeout=1):
        try:
            return self.queue.get(timeout=timeout)
        except Exception as e:
            return None

    def set_scheduling_job_id(self, scheduling_job_id):
        self.scheduling_job_id = scheduling_job_id



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os
import time

from notification_service.client import NotificationClient
from notification_service.server import NotificationServerRunner
from notification_service.util import db

_NS_DB_FILE = 'ns.db'
_NS_DB_URI = '%s%s' % ('sqlite:///', _NS_DB_FILE)
_NS_PORT = '50052'
_NS_URI = 'localhost:%s' % _NS_PORT


def start_notification_server():
    if os.path.exists(_NS_DB_FILE):
        os.remove(_NS_DB_FILE)
    config_file = os.path.dirname(os.path.dirname(__file__)) + '/notification_server.yaml'
    ns_server = NotificationServerRunner(config_file=config_file)
    db.create_all_tables(ns_server.config.db_uri)
    ns_server.start()
    wait_for_notification_server_started(_NS_URI)
    return ns_server


def stop_notification_server(ns_server):
    ns_server.stop()
    os.remove(_NS_DB_FILE)


def wait_for_notification_server_started(server_uri, timeout=60):
    start_time = time.monotonic()
    succeed = False
    while time.monotonic() < start_time + timeout:
        try:
            NotificationClient(server_uri)
            succeed = True
            break
        except Exception as e:
            print(e)
            time.sleep(0.1)
            continue
    if not succeed:
        raise Exception("Timeout to wait for notification server started.")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import os
import signal
import time

import psutil
from typing import List

logger = logging.getLogger(__name__)


def get_all_children_pids(current_pid=None) -> List:
    result = []
    if not psutil.pid_exists(current_pid):
        return result
    p = psutil.Process(current_pid)
    for pp in p.children():
        result.append(pp.pid)
        result.extend(get_all_children_pids(pp.pid))
    return result


def check_pid_exist(_pid):
    try:
        os.kill(_pid, 0)
    except OSError:
        return False
    else:
        return True


def stop_process(pid, process_name, timeout_sec: int = 60):
    """
    Try hard to kill the process with the given pid. It sends
    :param pid: The process pid to stop
    :param process_name: The process name to log
    :param timeout_sec: timeout before sending SIGKILL to kill the process
    """

    try:
        start_time = time.monotonic()
        while check_pid_exist(pid):
            if time.monotonic() - start_time > timeout_sec:
                raise RuntimeError(
                    "{} pid: {} does not exit after {} seconds.".format(process_name, pid, timeout_sec))
            os.kill(pid, signal.SIGTERM)
            time.sleep(1)
    except Exception:
        logger.warning("Failed to stop {} pid: {} with SIGTERM. Try to send SIGKILL".format(process_name, pid))
        try:
            os.kill(pid, signal.SIGKILL)
        except Exception as e:
            raise RuntimeError("Failed to kill {} pid: {} with SIGKILL.".format(process_name, pid)) from e

    logger.info("{} pid: {} stopped".format(process_name, pid))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from apscheduler.jobstores.base import BaseJobStore, JobLookupError, ConflictingIdError
from apscheduler.util import datetime_to_utc_timestamp, utc_timestamp_to_datetime
from apscheduler.job import Job
from airflow.models.periodic_task_model import PeriodicTaskModel
from airflow.utils.session import create_session

try:
    import cPickle as pickle
except ImportError:  # pragma: nocover
    import pickle

try:
    from sqlalchemy.exc import IntegrityError
    from sqlalchemy.sql.expression import null
except ImportError:  # pragma: nocover
    raise ImportError('SQLAlchemyJobStore requires SQLAlchemy installed')


class PeriodicTaskSQLAlchemyJobStore(BaseJobStore):
    """ The class saves the periodic scheduling information of periodic tasks to database. """
    def __init__(self, pickle_protocol=pickle.HIGHEST_PROTOCOL):
        super(PeriodicTaskSQLAlchemyJobStore, self).__init__()
        self.pickle_protocol = pickle_protocol
        self.session = None

    def set_session(self, session):
        """
        If you set an external session, after calling the interfaces(add_job, update_job, remove_job, remove_all_jobs),
        you need to call session.commit() to update to the database.
        """
        self.session = session

    def unset_session(self):
        self.session = None

    def lookup_job(self, job_id):
        def _internal_lookup_job(job_id, session):
            job_state = session.query(PeriodicTaskModel.job_state).filter(PeriodicTaskModel.id == job_id).scalar()
            return self._reconstitute_job(job_state) if job_state else None

        if self.session is None:
            with create_session() as session:
                return _internal_lookup_job(job_id, session)
        else:
            return _internal_lookup_job(job_id, self.session)

    def get_due_jobs(self, now):
        timestamp = datetime_to_utc_timestamp(now)
        return self._get_jobs(PeriodicTaskModel.next_run_time <= timestamp)

    def get_next_run_time(self):
        def _internal_get_next_run_time(session):
            next_run_time = session.query(PeriodicTaskModel.next_run_time).filter(
                PeriodicTaskModel.next_run_time != null()).order_by(PeriodicTaskModel.next_run_time).limit(1).scalar()
            return utc_timestamp_to_datetime(next_run_time)
        if self.session is None:
            with create_session() as session:
                return _internal_get_next_run_time(session)
        else:
            return _internal_get_next_run_time(self.session)

    def get_all_jobs(self):
        jobs = self._get_jobs()
        self._fix_paused_jobs_sorting(jobs)
        return jobs

    def add_job(self, job):
        r = PeriodicTaskModel()
        r.id = job.id
        r.next_run_time = datetime_to_utc_timestamp(job.next_run_time)
        r.job_state = pickle.dumps(job.__getstate__(), self.pickle_protocol)
        if self.session is None:
            with create_session() as session:
                try:
                    session.add(r)
                    session.commit()
                except IntegrityError:
                    raise ConflictingIdError(job.id)
        else:
            self.session.add(r)

    def update_job(self, job):
        def _internal_update_job(session):
            periodic_task_model = session.query(PeriodicTaskModel).filter(PeriodicTaskModel.id == job.id).first()
            if periodic_task_model is None:
                raise JobLookupError(job.id)
            else:
                periodic_task_model.next_run_time = datetime_to_utc_timestamp(job.next_run_time)
                periodic_task_model.job_state = pickle.dumps(job.__getstate__(), self.pickle_protocol)
                session.merge(periodic_task_model)
        if self.session is None:
            with create_session() as session:
                _internal_update_job(session)
        else:
            _internal_update_job(self.session)

    def remove_job(self, job_id):
        if self.session is None:
            with create_session() as session:
                session.query(PeriodicTaskModel).filter(PeriodicTaskModel.id == job_id).delete()
        else:
            self.session.query(PeriodicTaskModel).filter(PeriodicTaskModel.id == job_id).delete()

    def remove_all_jobs(self):
        if self.session is None:
            with create_session() as session:
                session.query(PeriodicTaskModel).delete()
        else:
            self.session.query(PeriodicTaskModel).delete()

    def _reconstitute_job(self, job_state):
        job_state = pickle.loads(job_state)
        job_state['jobstore'] = self
        job = Job.__new__(Job)
        job.__setstate__(job_state)
        job._scheduler = self._scheduler
        job._jobstore_alias = self._alias
        return job

    def _get_jobs(self, *conditions):
        if self.session is None:
            with create_session() as session:
                return self._internal_get_jobs(conditions, session)
        else:
            return self._internal_get_jobs(conditions, self.session)

    def _internal_get_jobs(self, conditions, session):
        jobs = []
        query = session.query(PeriodicTaskModel) \
            .order_by(PeriodicTaskModel.next_run_time)
        query = query.filter(*conditions) if conditions else query
        selectable_jobs = query.all()
        failed_job_ids = set()
        for job in selectable_jobs:
            try:
                jobs.append(self._reconstitute_job(job.job_state))
            except BaseException as e:
                self._logger.exception('Unable to restore job "%s exception [%s]" -- removing it', job.id, str(e))
                failed_job_ids.add(job.id)
        # Remove all the jobs we failed to restore
        if failed_job_ids:
            session.query(PeriodicTaskModel).filter(PeriodicTaskModel.id.in_(failed_job_ids)).delete()
            session.commit()
        return jobs



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os

import cloudpickle

from notification_service.base_notification import BaseEvent

from ai_flow.api.context_extractor import ContextExtractor, EventContext, Broadcast, ContextList

context_extractor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                      'context_extractor.pickle')


class EventKeyContextExtractor(ContextExtractor):
    def extract_context(self, event: BaseEvent) -> EventContext:
        if event.key == 'broadcast':
            return Broadcast()

        context_list = ContextList()
        context_list.add_context(event.key)
        return context_list


if __name__ == '__main__':
    e = EventKeyContextExtractor()

    with open(context_extractor_path, 'wb') as f:
        cloudpickle.dump(e, f)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add_periodic_task_model

Revision ID: cf3cb8473f3c
Revises: c12f471f01f4
Create Date: 2021-10-08 16:05:00.262365

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = 'cf3cb8473f3c'
down_revision = 'c12f471f01f4'
branch_labels = None
depends_on = None


def upgrade():
    """Apply add_periodic_task_model"""

    op.create_table(
        'periodic_task_model',
        sa.Column('id', sa.String(length=750), nullable=False),
        sa.Column('next_run_time', sa.Float(25), nullable=False),
        sa.Column('job_state', sa.LargeBinary, nullable=True),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    """Unapply add_periodic_task_model"""
    op.drop_table('periodic_task_model')




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add_ai_flow_tables

Revision ID: de1c96ef582a
Revises: 
Create Date: 2021-10-13 16:18:36.091074

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'de1c96ef582a'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('artifact',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.Column('artifact_type', sa.String(length=256), nullable=True),
                    sa.Column('description', sa.String(length=1000), nullable=True),
                    sa.Column('uri', sa.String(length=1000), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=True),
                    sa.Column('update_time', sa.BigInteger(), nullable=True),
                    sa.Column('properties', sa.String(length=1000), nullable=True),
                    sa.Column('is_deleted', sa.String(length=256), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('name'),
                    sqlite_autoincrement=True
                    )
    op.create_table('dataset',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.Column('format', sa.String(length=256), nullable=True),
                    sa.Column('description', sa.String(length=1000), nullable=True),
                    sa.Column('uri', sa.String(length=1000), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=True),
                    sa.Column('update_time', sa.BigInteger(), nullable=True),
                    sa.Column('properties', sa.String(length=1000), nullable=True),
                    sa.Column('name_list', sa.String(length=1000), nullable=True),
                    sa.Column('type_list', sa.String(length=1000), nullable=True),
                    sa.Column('catalog_name', sa.String(length=1000), nullable=True),
                    sa.Column('catalog_type', sa.String(length=1000), nullable=True),
                    sa.Column('catalog_database', sa.String(length=1000), nullable=True),
                    sa.Column('catalog_connection_uri', sa.String(length=1000), nullable=True),
                    sa.Column('catalog_table', sa.String(length=1000), nullable=True),
                    sa.Column('is_deleted', sa.String(length=256), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('name'),
                    sqlite_autoincrement=True
                    )
    op.create_table('event',
                    sa.Column('id', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('key', sa.String(length=1024), nullable=False),
                    sa.Column('version', sa.Integer(), nullable=False),
                    sa.Column('value', sa.String(length=1024), nullable=True),
                    sa.Column('event_type', sa.String(length=256), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sqlite_autoincrement=True
                    )
    op.create_table('metric_meta',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('metric_name', sa.String(length=256), nullable=False),
                    sa.Column('metric_type', sa.String(length=256), nullable=True),
                    sa.Column('metric_desc', sa.String(length=4096), nullable=True),
                    sa.Column('project_name', sa.String(length=256), nullable=False),
                    sa.Column('dataset_name', sa.String(length=256), nullable=True),
                    sa.Column('model_name', sa.String(length=256), nullable=True),
                    sa.Column('job_name', sa.String(length=256), nullable=True),
                    sa.Column('start_time', sa.BigInteger(), nullable=True),
                    sa.Column('end_time', sa.BigInteger(), nullable=True),
                    sa.Column('uri', sa.String(length=1024), nullable=True),
                    sa.Column('tags', sa.String(length=256), nullable=True),
                    sa.Column('properties', sa.String(length=1024), nullable=True),
                    sa.Column('is_deleted', sa.String(length=128), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('metric_name'),
                    sqlite_autoincrement=True
                    )
    op.create_table('project',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('properties', sa.String(length=1000), nullable=True),
                    sa.Column('uri', sa.String(length=1000), nullable=True),
                    sa.Column('is_deleted', sa.String(length=256), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('name'),
                    sqlite_autoincrement=True
                    )
    op.create_table('registered_model',
                    sa.Column('model_name', sa.String(length=255), nullable=False),
                    sa.Column('model_desc', sa.String(length=1000), nullable=True),
                    sa.PrimaryKeyConstraint('model_name', name='registered_model_pk'),
                    sa.UniqueConstraint('model_name')
                    )
    op.create_table('workflow_execution_event_handler_state',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('project_name', sa.String(length=256), nullable=False),
                    sa.Column('workflow_name', sa.String(length=256), nullable=False),
                    sa.Column('context', sa.String(length=256), nullable=False),
                    sa.Column('workflow_execution_id', sa.Text(), nullable=True),
                    sa.Column('state', sa.PickleType(), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('project_name', 'workflow_name', 'context'),
                    sqlite_autoincrement=True
                    )
    op.create_table('metric_summary',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('metric_name', sa.String(length=256), nullable=False),
                    sa.Column('metric_key', sa.String(length=256), nullable=False),
                    sa.Column('metric_value', sa.String(length=2048), nullable=False),
                    sa.Column('metric_timestamp', sa.BigInteger(), nullable=False),
                    sa.Column('model_version', sa.String(length=256), nullable=True),
                    sa.Column('job_execution_id', sa.String(length=256), nullable=True),
                    sa.Column('is_deleted', sa.String(length=128), nullable=True),
                    sa.ForeignKeyConstraint(['metric_name'], ['metric_meta.metric_name'], onupdate='cascade'),
                    sa.PrimaryKeyConstraint('uuid'),
                    sqlite_autoincrement=True
                    )
    op.create_table('model_relation',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.Column('project_id', sa.BigInteger(), nullable=True),
                    sa.Column('is_deleted', sa.String(length=256), nullable=True),
                    sa.ForeignKeyConstraint(['project_id'], ['project.uuid'], onupdate='cascade'),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('name'),
                    sqlite_autoincrement=True
                    )
    op.create_table('model_version',
                    sa.Column('model_name', sa.String(length=255), nullable=False),
                    sa.Column('model_version', sa.String(length=10), nullable=False),
                    sa.Column('model_path', sa.String(length=500), nullable=True),
                    sa.Column('model_type', sa.String(length=500), nullable=True),
                    sa.Column('version_desc', sa.String(length=1000), nullable=True),
                    sa.Column('version_status', sa.String(length=20), nullable=True),
                    sa.Column('current_stage', sa.String(length=20), nullable=False),
                    sa.ForeignKeyConstraint(['model_name'], ['registered_model.model_name'], onupdate='cascade',
                                            ondelete='cascade'),
                    sa.PrimaryKeyConstraint('model_name', 'model_version', 'current_stage', name='model_version_pk')
                    )
    op.create_table('project_snapshot',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('project_id', sa.BigInteger(), nullable=True),
                    sa.Column('signature', sa.String(length=255), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=True),
                    sa.ForeignKeyConstraint(['project_id'], ['project.uuid'], onupdate='cascade'),
                    sa.PrimaryKeyConstraint('uuid'),
                    sqlite_autoincrement=True
                    )
    op.create_table('workflow',
                    sa.Column('uuid', sa.BigInteger(), autoincrement=True, nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.Column('project_id', sa.BigInteger(), nullable=True),
                    sa.Column('properties', sa.String(length=1000), nullable=True),
                    sa.Column('create_time', sa.BigInteger(), nullable=True),
                    sa.Column('update_time', sa.BigInteger(), nullable=True),
                    sa.Column('is_deleted', sa.Boolean(), nullable=True),
                    sa.Column('context_extractor_in_bytes', sa.LargeBinary(), nullable=True),
                    sa.Column('graph', sa.Text(), nullable=True),
                    sa.Column('scheduling_rules', sa.Text(), nullable=True),
                    sa.Column('last_event_version', sa.BigInteger(), nullable=True),
                    sa.ForeignKeyConstraint(['project_id'], ['project.uuid'], onupdate='cascade'),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('project_id', 'name'),
                    sqlite_autoincrement=True
                    )
    op.create_table('model_version_relation',
                    sa.Column('version', sa.String(length=255), nullable=False),
                    sa.Column('model_id', sa.BigInteger(), nullable=False),
                    sa.Column('project_snapshot_id', sa.BigInteger(), nullable=True),
                    sa.Column('is_deleted', sa.String(length=256), nullable=True),
                    sa.ForeignKeyConstraint(['model_id'], ['model_relation.uuid'], onupdate='cascade'),
                    sa.ForeignKeyConstraint(['project_snapshot_id'], ['project_snapshot.uuid'], onupdate='cascade'),
                    sa.PrimaryKeyConstraint('version', 'model_id')
                    )


def downgrade():
    op.drop_table('model_version_relation')
    op.drop_table('workflow')
    op.drop_table('project_snapshot')
    op.drop_table('model_version')
    op.drop_table('model_relation')
    op.drop_table('metric_summary')
    op.drop_table('workflow_execution_event_handler_state')
    op.drop_table('registered_model')
    op.drop_table('project')
    op.drop_table('metric_meta')
    op.drop_table('event')
    op.drop_table('dataset')
    op.drop_table('artifact')
